import dataclasses
import functools
import logging
import typing as t

import a2.dataset
import a2.plotting.analysis
import a2.training.tracking
import datasets
import numpy as np
import sklearn.model_selection
import transformers
import xarray


@dataclasses.dataclass
class HyperParametersDebertaClassifier:
    """
    Hold hyper parameters for Hugging Face DeBERTa classifier
    """

    learning_rate: float = 3e-05
    batch_size: int = 32
    weight_decay: float = 0.01
    epochs: int = 1
    warmup_ratio: float = 0.4480456499617466
    hidden_dropout_prob: float = 0.1
    attention_probs_dropout_prob: float = 0.1
    cls_dropout: float = 0.1
    lr_scheduler_type: str = "linear"


def _compute_metrics(eval_pred):
    """
    Adopt Hugging Face metric output to prefered metric
    (f1-score) for this project

    Parameters:
    ----------
    eval_pred: Predictions and labels of Hugging Face model while training

    Returns
    -------
    dictionary of metrics
    """
    predictions, labels = eval_pred
    predictions = predictions.argmax(axis=-1)
    classification_report = a2.plotting.analysis.classification_report(
        labels,
        predictions,
        target_names=["not raining", "raining"],
        output_dict=True,
    )
    f1_weighted_average = classification_report["weighted avg"]["f1-score"]
    f1_macro_average = classification_report["macro avg"]["f1-score"]
    f1_not_raining = classification_report["not raining"]["f1-score"]
    f1_raining = classification_report["raining"]["f1-score"]
    return {
        "f1_not_raining": f1_not_raining,
        "f1_raining": f1_raining,
        "f1_weighted_average": f1_weighted_average,
        "f1_macro_average": f1_macro_average,
    }


class HuggingFaceTrainerClass:
    """
    Used to train Hugging Face models
    """

    def __init__(
        self,
        model_folder: str,
        num_labels: int = 2,
        config: t.Optional[t.Dict] = None,
    ):
        self.model_folder = model_folder
        self.num_labels = num_labels
        if config is None:
            self.db_config_base = transformers.AutoConfig.from_pretrained(model_folder, num_labels=num_labels)
        else:
            self.db_config_base = config
        self.hyper_parameters = HyperParametersDebertaClassifier()

    def get_model(self, params: t.Dict, mantik: bool = True, base_model_trainable: bool = True):
        db_config = self.db_config_base
        if params is not None:
            db_config.update({"cls_dropout": params["cls_dropout"]})
        db_config.update({"num_labels": self.num_labels})
        model = transformers.AutoModelForSequenceClassification.from_pretrained(self.model_folder, config=db_config)
        if not base_model_trainable:
            for param in model.base_model.parameters():
                param.requires_grad = False
        if mantik:
            a2.training.tracking.initialize_mantik()
        return model

    def get_trainer(
        self,
        dataset: t.Union[datasets.Dataset, datasets.DatasetDict],
        hyper_parameters: HyperParametersDebertaClassifier = None,
        tokenizer: t.Optional[transformers.DebertaTokenizer] = None,
        folder_output: str = "output/",
        hyper_tuning: bool = False,
        fp16: bool = True,
        evaluate: bool = False,
        mantik: bool = True,
        disable_tqdm: bool = False,
        callbacks: list | None = None,
        base_model_trainable: bool = True,
        trainer_class=transformers.Trainer,
        logging_steps: int = 500,
        evaluation_strategy: str = "steps",
        save_strategy: str = "epoch",
        load_best_model_at_end: bool = True,
    ):
        """
        Returns Hugging Face trainer object

        Parameters:
        ----------
        dataset: Dataset in Hugging Face format
        hyper_parameters: hyper parameters in the form of data
                          class `HyperParametersDebertaClassifier`
        tokenizer: Hugging Face tokenizer
        folder_output: Folder to save training outputs
        hyper_tuning: Whether trainer used for hyper tuning
        fp16: Whether to use fp16 16-bit (mixed) precision training
              instead of 32-bit training.
        evaluate: Whether trainer only used for evaluation
        mantik: Whether using mantik for tracking
        logging_steps: Number of steps before hugging face prints
        logging_steps: When to evaluate, after "steps" or "epoch"

        Returns
        -------
        Hugging Face Trainer
        """
        if hyper_parameters is None:
            hyper_parameters = HyperParametersDebertaClassifier()
        self.hyper_parameters = hyper_parameters
        if tokenizer is None:
            tokenizer = transformers.AutoTokenizer.from_pretrained(self.model_folder)
        if load_best_model_at_end and save_strategy != evaluation_strategy:
            logging.info(
                f"Setting {save_strategy=} equal to {evaluation_strategy=} "
                f"(required when {load_best_model_at_end=} = True)"
            )
            save_strategy = evaluation_strategy
        if not hyper_tuning:
            args = transformers.TrainingArguments(
                folder_output,
                learning_rate=hyper_parameters.learning_rate,
                warmup_ratio=hyper_parameters.warmup_ratio,
                lr_scheduler_type=hyper_parameters.lr_scheduler_type,
                disable_tqdm=disable_tqdm,
                fp16=fp16,
                evaluation_strategy=evaluation_strategy,
                per_device_train_batch_size=hyper_parameters.batch_size,
                per_device_eval_batch_size=hyper_parameters.batch_size,
                num_train_epochs=hyper_parameters.epochs,
                weight_decay=hyper_parameters.weight_decay,
                report_to=None,
                save_strategy=save_strategy,
                load_best_model_at_end=load_best_model_at_end,
                logging_steps=logging_steps,
            )
        else:
            args = transformers.TrainingArguments(
                folder_output,
                disable_tqdm=True,
                fp16=fp16,
                evaluation_strategy=evaluation_strategy,
                report_to=None,
                save_strategy=save_strategy,
                load_best_model_at_end=load_best_model_at_end,
            )
        model_init = functools.partial(self.get_model, mantik=mantik, base_model_trainable=base_model_trainable)
        if evaluate:
            return trainer_class(
                model_init=model_init,
                args=args,
                tokenizer=tokenizer,
                compute_metrics=_compute_metrics,
            )
        return trainer_class(
            model_init=model_init,
            args=args,
            train_dataset=dataset["train"],
            eval_dataset=dataset["test"],
            tokenizer=tokenizer,
            compute_metrics=_compute_metrics,
            callbacks=callbacks,
        )


def split_training_set(
    ds: xarray.Dataset,
    key_stratify: str = "raining",
    test_size: float = 0.2,
    random_state: int = 42,
    shuffle: bool = True,
):
    """
    Returns indices of training and validation set

    Parameters:
    ----------
    ds: Xarray dataset
    key_stratify: Stratified based on this key, `None` if not required
    test_size: Fraction of validation set [0-1]
    random_state: Random seed to initialize selection
    shuffle: Whether or not to shuffle the data before splitting.
             If shuffle=False then stratify must be None.

    Returns
    -------
    Indices of training and test set
    """
    if key_stratify is None:
        stratify = ds[key_stratify].values
    else:
        stratify = None
    indices_train, indices_validate = sklearn.model_selection.train_test_split(
        np.arange(ds[a2.dataset.utils_dataset.get_variable_name_first(ds)].shape[0]),
        test_size=test_size,
        random_state=random_state,
        shuffle=shuffle,
        stratify=stratify,
    )
    return indices_train, indices_validate



import io
import pathlib
import sys
import typing as t

import a2.dataset
import a2.utils.file_handling
import numpy as np
import pandas as pd
import urllib3
import xarray


class IOCapture:
    """
    Used to capture terminal output as string.

    Usually only used for testing.
    """

    def __init__(self):
        self.captured_output = io.StringIO()
        sys.stdout = self.captured_output

    def stop(self) -> None:
        """
        Resets redirect
        """
        sys.stdout = sys.__stdout__

    def return_capture_stop(self) -> str:
        """
        Resets redirect and returns captured value as string
        """
        self.stop()
        return self.captured_output.getvalue()


def ordered(obj: object):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    elif isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


def json_equal(json1: object, json2: object) -> bool:
    """
    Assert if two json objects are equal

    Prints out more useful error messages if not the case.
    Treats empty string as nan during comparison.
    Parameters:
    ----------
    json1: First json object
    json2: Second json object

    Returns
    -------
    Bool
    """
    json1_ordered = ordered(json1)
    json2_ordered = ordered(json2)
    if json1_ordered != json2_ordered:
        if type(json1_ordered) != type(json2_ordered):
            raise AssertionError(
                f"json1_ordered: {json1_ordered} not the same type: "
                "{type(json1_ordered)} as json2_ordered: {json2_ordered} "
                "of type {type(json2_ordered)}"
            )
        if len(json1_ordered) != len(json2_ordered):
            raise AssertionError(
                f"len(json1_ordered): {len(json1_ordered)} " "!= len(json2_ordered): {len(json2_ordered)}"
            )
        if isinstance(json1_ordered, list):
            for item1, item2 in zip(json1_ordered, json2_ordered):
                json_equal(item1, item2)
        if json1_ordered != json2_ordered:
            raise AssertionError("Json (element) not equal: " f"{json1_ordered} \n != \n {json2_ordered}")
    return True


def assert_presence_variables(ds: xarray.Dataset, variables: t.Sequence[str]):
    for var in variables:
        if var not in ds.variables.keys():
            raise ValueError(f"{var} not available in ds: {ds}")


def print_text_as_csv(text: str, filename: t.Union[str, pathlib.Path] = "/tmp/tmp.txt"):
    a2.utils.file_handling.csv_create(filename, row=[text], check_is_file_or_empty=False)
    print(open(filename).read())


def check_internet_connection():
    http = urllib3.PoolManager(timeout=3.0)
    try:
        r = http.request("GET", "google.com", preload_content=False)
    except urllib3.exceptions.MaxRetryError:
        return False
    code = r.status
    r.release_conn()
    if code == 200:
        return True
    else:
        return False


def check_python_version():
    from platform import python_version

    return python_version()


def print_copy_paste_fake_dataset(ds, name="fake_dataset"):
    print(f"def {name}():")

    def print_array(values, dtype):
        if "<" in dtype.__str__():
            dtype = "str_"
        elif "object" == dtype.__str__():
            dtype = "object_"
        elif "bool" == dtype.__str__():
            dtype = "bool_"
        elif "datetime64[ns]" == dtype.__str__():
            dtype = "np.datetime64"
        return (
            "np.array(["
            + ", ".join(
                [
                    "np.nan"
                    if not isinstance(x, str) and not isinstance(x, dict) and np.isnan(x)
                    else x.__str__()
                    if not isinstance(x, str) and not isinstance(x, np.datetime64)
                    else f"'{x}'"
                    for x in values
                ]
            )
            + f"], dtype=np.{dtype})"
        )

    for k, v in ds.variables.items():
        value_array = np.array(v.values, dtype=v.values.dtype)
        print(f"{k} = {print_array(value_array, dtype=v.values.dtype)}")
    print("return xarray.Dataset(")
    print("\tdata_vars=dict(")
    for k, v in ds.variables.items():
        if k == "index":
            continue
        print(f'\t\t{k}=(["index"], {k}),')
    print("\t),")
    print("\tcoords=dict(index=index),")
    print(")")


def print_debug(x):
    print(f"{np.min(x)=}")
    print(f"{np.max(x)=}")
    print(f"{np.mean(x)=}")
    print(f"{np.isnan(x).sum()=}")
    print(f"{x[:10]=}")


def not_overlapping(left: list, right: list) -> list:
    return list(set(left) - set(right))


def assert_equal_pandas_dataframe(left: pd.DataFrame, right: pd.DataFrame) -> None:
    try:
        pd.testing.assert_frame_equal(left, right, check_like=True)
    except AssertionError as e:
        message = ""
        if not_overlapping(left.columns, right.columns):
            message += f"df2 missing {not_overlapping(left.columns, right.columns)}"
        if not_overlapping(right.columns, left.columns):
            message += f"df1 missing {not_overlapping(right.columns, left.columns)}"
        raise AssertionError(f"{message}\n{e}")



import logging
import pathlib
import typing as t

import a2.dataset
import a2.plotting.utils_plotting
import a2.utils.constants
import matplotlib.offsetbox
import matplotlib.pyplot as plt
import numpy as np
import xarray


FILE_LOC = pathlib.Path(__file__).parent
DATA_FOLDER = FILE_LOC / "../data"


def plot_histogram_2d(
    x: t.Union[np.ndarray, str],
    y: t.Union[np.ndarray, str],
    ds: xarray.Dataset | None = None,
    bins: t.Sequence | None = None,
    xlim: t.Sequence | None = None,
    ylim: t.Sequence | None = None,
    ax: a2.utils.constants.TYPE_MATPLOTLIB_AXES | None = None,
    ax_colorbar: a2.utils.constants.TYPE_MATPLOTLIB_AXES | None = None,
    log: t.Union[bool, t.List[bool]] = False,
    filename: t.Union[str, pathlib.Path] = None,
    fig: a2.utils.constants.TYPE_MATPLOTLIB_FIGURES | None = None,
    n_bins: t.Union[int, t.List[int]] = 60,
    norm: str | None = None,
    linear_thresh: t.Union[float, t.List[float]] = 1e-9,
    label_x: str | None = None,
    label_y: str | None = None,
    label_colorbar: str | None = None,
    font_size: int = 12,
    overplot_values: bool = False,
    cmap: str = "plasma",
    overplot_round_base: int | None = None,
    overplot_color: str = "gray",
    return_matrix: bool = False,
    aspect: str = "auto",
    vmin: float | None = None,
    vmax: float | None = None,
    marginal_x: str = "histogram",
    marginal_x_label_x: bool = False,
    marginal_x_label_y: str = "N",
    marginal_x_show_xticks: bool = False,
    marginal_y: str = "histogram",
    marginal_y_label_x: str = "N",
    marginal_y_label_y: bool = False,
    marginal_y_show_yticks: bool = False,
    marginal_color: str | None = None,
    figure_size: t.List[float] | None = None,
    colormap: str = "viridis",
):
    return _plot_histogram_2d(
        x,
        y,
        ds=ds,
        bins=bins,
        xlim=xlim,
        ylim=ylim,
        ax=ax,
        ax_colorbar=ax_colorbar,
        log=log,
        filename=filename,
        fig=fig,
        n_bins=n_bins,
        norm=norm,
        linear_thresh=linear_thresh,
        label_x=label_x,
        label_y=label_y,
        label_colorbar=label_colorbar,
        font_size=font_size,
        overplot_values=overplot_values,
        cmap=cmap,
        overplot_round_base=overplot_round_base,
        overplot_color=overplot_color,
        return_matrix=return_matrix,
        aspect=aspect,
        vmin=vmin,
        vmax=vmax,
        marginal_x=marginal_x,
        marginal_x_label_x=marginal_x_label_x,
        marginal_x_label_y=marginal_x_label_y,
        marginal_x_show_xticks=marginal_x_show_xticks,
        marginal_y=marginal_y,
        marginal_y_label_x=marginal_y_label_x,
        marginal_y_label_y=marginal_y_label_y,
        marginal_y_show_yticks=marginal_y_show_yticks,
        marginal_color=marginal_color,
        figure_size=figure_size,
        colormap=colormap,
    )


def _plot_histogram_2d(
    x: t.Union[np.ndarray, str],
    y: t.Union[np.ndarray, str],
    ds: xarray.Dataset | None = None,
    bins: t.Sequence | None = None,
    xlim: t.Sequence | None = None,
    ylim: t.Sequence | None = None,
    ax: a2.utils.constants.TYPE_MATPLOTLIB_AXES | None = None,
    ax_colorbar: a2.utils.constants.TYPE_MATPLOTLIB_AXES | None = None,
    log: t.Union[bool, t.List[bool]] = False,
    filename: t.Union[str, pathlib.Path] = None,
    fig: a2.utils.constants.TYPE_MATPLOTLIB_FIGURES | None = None,
    n_bins: t.Union[int, t.List[int]] = 60,
    norm: str | None = None,
    linear_thresh: t.Union[float, t.List[float]] = 1e-9,
    label_x: str | None = None,
    label_y: str | None = None,
    label_colorbar: str | None = None,
    font_size: int = 12,
    overplot_values: bool = False,
    cmap: str = "plasma",
    overplot_round_base: int | None = None,
    overplot_color: str = "gray",
    return_matrix: bool = False,
    aspect: str = "auto",
    vmin: float | None = None,
    vmax: float | None = None,
    marginal_x: str = "histogram",
    marginal_x_label_x: bool = False,
    marginal_x_label_y: str = "N",
    marginal_x_show_xticks: bool = False,
    marginal_y: str = "histogram",
    marginal_y_label_x: str = "N",
    marginal_y_label_y: bool = False,
    marginal_y_show_yticks: bool = False,
    marginal_color: str | None = None,
    figure_size: t.List[float] | None = None,
    colormap: str = "viridis",
) -> t.Union[plt.axes, t.Tuple[plt.axes, t.Sequence]]:
    """
    plots 2d histogram

    Log types included `False`, `log`/`True`, `symlog`.
    Norm takes care of colorbar scale, so far included: `log`
    Parameters:
    ----------
    x: values binned on x-axis
    y: values binned on y-axis
    bins: if None computed based on data provided, otherwise should
          provide [x_edges, y_edges]
    xlim: limits of x-axis and bin_edges, if None determined from x
    ylim: limits of y-axis and bin_edges, if None determined from y
    ax: matplotlib axes
    log: type of bins
    filename: plot saved to file if provided
    fig: matplotlib figure
    n_bins: number of bins
    norm: scaling of colorbar
    linear_thresh: required if using log='symlog' to indicate
                   where scale turns linear
    label_x: label of x-axis
    label_y: label of y-axis
    font_size: size of font
    overplot_values: show number of samples on plot
    overplot_round_base: Round values to be overplotted to this base int
    return_matrix: Additionally return histogram values
    marginal_x: Add additional plot on top of histogram plot ("histogram"/None)
    marginal_x_label_x: Label along x-axis for plot `marginal_x`
    marginal_x_label_y: Label along y-axis for plot `marginal_x`
    marginal_x_show_xticks: Wether to show ticks along the x-axis for plot `marginal_x`
    marginal_y: Add additional plot on top of histogram plot ("histogram"/None)
    marginal_y_label_x: Label along x-axis for plot `marginal_y`
    marginal_y_label_y: Label along y-axis for plot `marginal_y`
    marginal_y_show_yticks: Wether to show ticks along the y-axis for plot `marginal_y`
    figure_size: Size of figure (height, width), default (10, 6)
    colormap: Matplotlib colormap name
    vmin: Minimum value of colormap
    vmax: Maximum value of colormap

    Returns
    -------
    axes
    """

    if marginal_x is not None and marginal_y is not None:
        fig, axes, axes_colorbar = a2.plotting.utils_plotting.create_axes_grid(
            n_cols=2,
            n_rows=2,
            figure_size=figure_size,
            skip_row_col=[[0, 1]],
            colorbar_width=0.02,
            spacing_x=0.03,
            spacing_y=0.03,
            widths_along_x=[0.2, 0.1],
            heights_along_y=[0.1, 0.2],
            colorbar_include_row_col=[[1, 1]],
            spacing_colorbar=0.03,
        )
        ax = axes[1][0]
        ax_colorbar = axes_colorbar[1][1]
    else:
        fig, ax = a2.plotting.utils_plotting.create_figure_axes(fig=fig, ax=ax, font_size=font_size)

    def get_label(label):
        if label is not None and label:
            return label
        else:
            return ""

    if isinstance(x, str):
        if ds is not None:
            if label_x is None:
                label_x = x
            x = ds[x].values
        else:
            ValueError(f"{ds=}, if {x=} given as string, dataset required")
    if isinstance(y, str):
        if ds is not None:
            if label_y is None:
                label_y = y
            y = ds[y].values
        else:
            ValueError(f"{ds=}, if {y=} given as string, dataset required")

    x, y = _get_xy_values(x, y)

    xlim, ylim, log, n_bins, linear_thresh = _prepare_parameters(xlim, ylim, log, n_bins, linear_thresh)

    if bins is None:
        bin_edges_x = get_bin_edges(
            data=x,
            n_bins=n_bins[0],
            linear_thresh=linear_thresh[0],
            log=log[0],
            vmin=xlim[0],
            vmax=xlim[1],
        )
        bin_edges_y = get_bin_edges(
            data=y,
            n_bins=n_bins[1],
            linear_thresh=linear_thresh[1],
            log=log[1],
            vmin=ylim[0],
            vmax=ylim[1],
        )
    else:
        bin_edges_x, bin_edges_y = bins
    a2.utils.checks.validate_array(bin_edges_x)  # type: ignore
    a2.utils.checks.validate_array(bin_edges_y)  # type: ignore

    norm_object = a2.plotting.utils_plotting.get_norm(norm, vmin=vmin, vmax=vmax)
    H, bin_edges_x, bin_edges_y = np.histogram2d(x, y, bins=[np.array(bin_edges_x), np.array(bin_edges_y)])
    H_plot = H.T
    X, Y = np.meshgrid(bin_edges_x, bin_edges_y)
    plot = ax.pcolormesh(X, Y, H_plot, norm=norm_object, cmap=colormap)
    if overplot_values:
        a2.plotting.utils_plotting.overplot_values(
            H, ax, len(bin_edges_x) - 1, len(bin_edges_y) - 1, color=overplot_color, round_to_base=overplot_round_base
        )
    colorbar = plt.colorbar(plot, cax=ax_colorbar)
    ax_colorbar = colorbar.ax
    if label_colorbar is not None:
        ax_colorbar.set_ylabel(label_colorbar)

    if xlim[0] is None:
        xlim[0] = bin_edges_x[0]
    if xlim[1] is None:
        xlim[1] = bin_edges_x[-1]
    if ylim[0] is None:
        ylim[0] = bin_edges_y[0]
    if ylim[1] is None:
        ylim[1] = bin_edges_y[-1]

    a2.plotting.utils_plotting.set_x_log(ax, log[0], linear_thresh=linear_thresh[0])
    a2.plotting.utils_plotting.set_y_log(ax, log[1], linear_thresh=linear_thresh[1])
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    if label_x is not None:
        ax.set_xlabel(label_x)
    if label_y is not None:
        ax.set_ylabel(label_y)

    if marginal_x == "histogram":
        plot_histogram(
            x,
            bin_edges=bin_edges_x,
            xlim=xlim,
            ylim=None,
            ax=axes[0][0],
            log=log[0],
            fig=fig,
            n_bins=n_bins[0],
            linear_thresh=linear_thresh[0],
            label_x=get_label(marginal_x_label_x),
            label_y=get_label(marginal_x_label_y),
            font_size=font_size,
            return_plot=False,
            color=marginal_color,
        )
        if not marginal_x_show_xticks:
            a2.plotting.utils_plotting.remove_tick_labels(axes[0][0], "x")
    if marginal_y == "histogram":
        plot_histogram(
            y,
            bin_edges=bin_edges_y,
            xlim=None,
            ylim=ylim,
            ax=axes[1][1],
            log=log[1],
            fig=fig,
            n_bins=n_bins[1],
            linear_thresh=linear_thresh[1],
            label_x=get_label(marginal_y_label_x),
            label_y=get_label(marginal_y_label_y),
            font_size=font_size,
            vertical=True,
            color=marginal_color,
        )
        if not marginal_y_show_yticks:
            a2.plotting.utils_plotting.remove_tick_labels(axes[1][1], "y")
    if fig is not None and marginal_x is None and marginal_y is None:
        fig.tight_layout()
    a2.plotting.utils_plotting.save_figure(fig, filename)
    if return_matrix:
        return ax, H
    return ax


def _prepare_parameters(xlim, ylim, log, n_bins, linear_thresh):
    log = a2.plotting.utils_plotting.to_list(log)
    n_bins = a2.plotting.utils_plotting.to_list(n_bins)
    xlim = a2.plotting.utils_plotting.to_list(xlim)
    ylim = a2.plotting.utils_plotting.to_list(ylim)
    n_bins = [i + 1 if i is not None else i for i in n_bins]  # using bin edges later, where n_edges = n_bins + 1
    linear_thresh = a2.plotting.utils_plotting.to_list(linear_thresh)
    return xlim, ylim, log, n_bins, linear_thresh


def _get_xy_values(x, y):
    x = np.ndarray.flatten(x)  # type: ignore
    a2.utils.checks.validate_array(x)
    y = np.ndarray.flatten(y)  # type: ignore
    a2.utils.checks.validate_array(y)
    if x.shape != y.shape:
        raise Exception(f"x and y need to be of same shape: {np.shape(x)} != {np.shape(y)}")
    return x, y


def annotate_histogram(
    ax: plt.axes, plot, labels: list, fontproperties=None, as_label: str | None = None, fontsize: int = 8
):
    """
    annotate labels to individual bar of histogram plot
    To download emoji images refer to `download_emoji` in
    notebooks/dataset/visualize_raw_tweets.ipynb

    Parameters:
    ----------
    ax: matplotlib axes
    plot: matplotlib plot object
    labels: List of labels for individual bars
    fontproperties: matplotlib fontproperties
    as_label: Annotate the labels of an axis instead of the bars directly
    font_size: size of font

    Returns
    -------
    """
    xycoords = "data"
    zoom = 0.5
    verticalalignment = "bottom"
    for num_label, (rect1, label) in enumerate(zip(plot, labels)):
        height = rect1.get_height()
        x_y = (rect1.get_x() + rect1.get_width() / 2, height + 5)
        x_y = [int(i) for i in x_y]
        x_y_images = x_y
        if as_label == "x":
            xycoords = "axes fraction"
            zoom = 0.3
            spacing_labels = 1 / len(labels)
            x_y = [spacing_labels / 2 + num_label * spacing_labels, -0.03]
            x_y_images = x_y[:]
            ax.tick_params(labelbottom=False)
            verticalalignment = "top"
            x_y_images[1] -= 0.05
        if a2.dataset.emojis.is_emoji(label):
            png = a2.dataset.emojis.get_emoji_filename(label)
            try:
                emoj = plt.imread(f"{DATA_FOLDER.__str__()}/emoji/emoji_images/{png}")
            except FileNotFoundError:
                png_alternative = png.replace("_fe0f", "")
                emoj = plt.imread(f"{DATA_FOLDER.__str__()}/emoji/emoji_images/{png_alternative}")
            imagebox = matplotlib.offsetbox.OffsetImage(emoj, zoom=zoom)
            ab = matplotlib.offsetbox.AnnotationBbox(imagebox, x_y_images, frameon=False, xycoords=xycoords)
            ax.add_artist(ab)
        else:
            ax.annotate(
                label,
                x_y,
                ha="center",
                va=verticalalignment,
                fontsize=fontsize,
                fontproperties=fontproperties,
                rotation=90,
                xycoords=xycoords,
            )


def plot_histogram(
    x: np.ndarray,
    ds=None,
    bin_edges=None,
    xlim: t.Optional[t.Sequence] = None,
    ylim: t.Optional[t.Sequence] = None,
    ax: t.Optional[plt.axes] = None,
    log: t.Union[bool, t.List[bool]] = False,
    filename: t.Union[str, pathlib.Path] = None,
    fig: t.Optional[plt.figure] = None,
    n_bins: int = 60,
    linear_thresh: t.Optional[float] = None,
    label_x: t.Optional[str] = None,
    label_y: t.Optional[str] = None,
    font_size: int = 12,
    label: str = None,
    return_plot: bool = False,
    vertical: bool = False,
    alpha: float = 1,
    annotatations_bars: t.Optional[list] = None,
    color: str | None = None,
    min_counts: int = 0,
) -> t.Union[plt.axes, t.Tuple[plt.figure, plt.axes]]:
    """
    plots 1d histogram

    Log types included `False`, `log`/`True`, `symlog`. Norm takes care of
    colorbar scale, so far included: `log`
    Parameters:
    ----------
    x: values binned on x-axis
    bin_edges: if None computed based on data provided, otherwise should
               provide [x_edges, y_edges]
    xlim: limits of x-axis and bin_edges, if None determined from x
    ylim: limits of y-axis and bin_edges, if None determined from y
    ax: matplotlib axes
    log: type of bins
    filename: plot saved to file if provided
    fig: matplotlib figure
    n_bins: number of bins
    norm: scaling of colorbar
    linear_thresh: required if using log='symlog' to indicate where
                   scale turns linear
    label_x: label of x-axis
    label_y: label of y-axis
    font_size: size of font
    alpha: alpha value of plot
    label: Label of plot
    return_plot: return axes and bar plot object

    Returns
    -------
    axes
    """
    if all([isinstance(i, str) for i in x]):
        labels_bars, x, counts = np.unique(x, return_inverse=True, return_counts=True)
        n_bins = len(labels_bars)
        annotatations_bars = [_la if c >= min_counts else "" for _la, c in zip(labels_bars, counts)]
        xlim = [0, np.max(x)]
    log = a2.plotting.utils_plotting.to_list(log)
    xlim = a2.plotting.utils_plotting.to_list(xlim)
    ylim = a2.plotting.utils_plotting.to_list(ylim)

    if ds is not None and isinstance(x, str):
        if label_x is None:
            label_x = x
        x = ds[x].values
    x = flatten_array(x)

    fig, ax = a2.plotting.utils_plotting.create_figure_axes(fig=fig, ax=ax, font_size=font_size)
    if min_counts > 0:
        ax.axhline(min_counts)

    if bin_edges is None:
        bin_edges, linear_thresh = get_bin_edges(
            data=x,
            n_bins=n_bins,
            linear_thresh=linear_thresh,
            log=log[0],
            return_linear_thresh=True,
            vmin=xlim[0],
            vmax=xlim[1],
        )
    (
        hist,
        bin_edges,
    ) = np.histogram(x, bins=bin_edges)
    bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2.0

    plot = plot_bar(
        bin_centers,
        hist,
        width_bars=np.diff(bin_edges),
        xlim=xlim,
        ylim=ylim,
        ax=ax,
        log=log,
        linear_thresh=linear_thresh,
        label_x=label_x,
        label_y=label_y,
        vertical=vertical,
        alpha=alpha,
        color=color,
    )
    if annotatations_bars is not None:
        annotate_histogram(ax, plot, labels=annotatations_bars, as_label="x", fontsize=font_size)
    fig.tight_layout()
    a2.plotting.utils_plotting.save_figure(fig, filename)
    print(f"{log=}")
    if return_plot:
        return ax, plot
    return ax


def plot_bar(
    bin_centers: np.ndarray,
    hist: np.ndarray,
    width_bars: np.ndarray | None = None,
    xlim: t.Optional[t.Sequence] = None,
    ylim: t.Optional[t.Sequence] = None,
    ax: t.Optional[plt.axes] = None,
    fig: a2.utils.constants.TYPE_MATPLOTLIB_FIGURES | None = None,
    log: t.Union[bool, t.List[bool]] = False,
    linear_thresh: t.Optional[float] = None,
    label_x: t.Optional[str] = None,
    label_y: t.Optional[str] = None,
    vertical: bool = False,
    alpha: float = 1,
    font_size: int = 8,
    replace_x_labels_at: t.Sequence | None = None,
    replace_x_labels_with: t.Sequence | None = None,
    replace_y_labels_at: t.Sequence | None = None,
    replace_y_labels_with: t.Sequence | None = None,
    color: str | None = None,
):
    """
    plots 1d bar plot

    Parameters:
    ----------
    bin_centers: Center position of bars
    hist: Bar values
    width_bars: Bar widths
    xlim: limits of x-axis and bin_edges, if None determined from x
    ylim: limits of y-axis and bin_edges, if None determined from y
    ax: matplotlib axes
    fig: matplotlib figure
    log: type of bins
    linear_thresh: required if using log='symlog' to indicate where
    label_x: label of x-axis
    label_y: label of y-axis
    Vertical: Plot bar plot along vertical axis instead of default horizontal
    alpha: Alpha value of bars
    font_size: Size of font
    replace_x_labels_at: Tick values of x-axis to be replaced
    replace_x_labels_with: Replacement values for tick values of x-axis
    replace_y_labels_at: Tick values of y-axis to be replaced
    replace_y_labels_with: Replacement values for tick values of y-axis
    color: Color of bars

    Returns
    -------
    plot
    """
    log = a2.plotting.utils_plotting.to_list(log)
    xlim = a2.plotting.utils_plotting.to_list(xlim)
    ylim = a2.plotting.utils_plotting.to_list(ylim)

    if ax is None:
        fig, ax = a2.plotting.utils_plotting.create_figure_axes(fig=fig, ax=ax, font_size=font_size)
    if vertical:
        plot = ax.barh(bin_centers, hist, height=width_bars, edgecolor="black", alpha=alpha, color=color)
    else:
        plot = ax.bar(bin_centers, hist, width=width_bars, edgecolor="black", alpha=alpha, color=color)

    a2.plotting.utils_plotting.set_x_log(ax, log[0], linear_thresh=linear_thresh)
    a2.plotting.utils_plotting.set_y_log(ax, log[1], linear_thresh=linear_thresh)
    a2.plotting.utils_plotting.set_axis_tick_labels(ax, replace_x_labels_at, replace_x_labels_with, axis="x")
    a2.plotting.utils_plotting.set_axis_tick_labels(ax, replace_y_labels_at, replace_y_labels_with, axis="y")
    if label_x is not None:
        ax.set_xlabel(label_x)
    if label_y is not None:
        ax.set_ylabel(label_y)
    ax.set_ylim(ylim)
    ax.set_xlim(xlim)
    return plot


def flatten_array(x):
    x = np.asarray(x)
    x = np.ndarray.flatten(x)
    return x


def get_bin_edges(
    vmin: t.Optional[float] = None,
    vmax: t.Optional[float] = None,
    linear_thresh: t.Optional[float] = None,
    n_bins: int = 60,
    data: t.Optional[np.ndarray] = None,
    log: t.Union[str, bool] = False,
    return_linear_thresh: bool = False,
) -> t.Union[t.Tuple[np.ndarray, t.Optional[float]], np.ndarray]:
    """
    returns bin edges for plots

    Log types included `False`, `log`/`True`, `symlog`
    Parameters:
    ----------
    vmin: minimum value of data
    vmax: maximum value of data
    linear_thresh: threshold below which bins are linear to include zero values
    n_bins: number of bins for logarithmic part of bins
    data: if provided used to compute `vmin` and `vmax`
    log: type of bins

    Returns
    -------
    bin edges
    """
    if data is not None and vmin is None:
        vmin = data.min()
    if data is not None and vmax is None:
        vmax = data.max()
    if vmin is None or vmax is None:
        raise Exception(f"Need to specify vmin {vmin} and {vmax} or provide data: {data}!")
    if not log:
        bins = np.linspace(vmin, vmax, n_bins)
    elif log == "symlog":
        if linear_thresh is None:
            abs_max = abs(vmax)
            abs_min = abs(vmin)
            linear_thresh = abs_min if abs_min < abs_max or abs_min == 0 else abs_max if abs_max != 0 else abs_min
            logging.info(f"Setting: linear_thresh: {linear_thresh} with vmin: {vmin}" " and vmax: {vmax}!")
        bins = _get_bin_edges_symlog(vmin, vmax, linear_thresh, n_bins=n_bins)
    else:
        bins = 10 ** np.linspace(np.log10(vmin), np.log10(vmax), n_bins)

    if return_linear_thresh:
        return bins, linear_thresh
    else:
        return bins


def _get_bin_edges_symlog(
    vmin: float,
    vmax: float,
    linear_thresh: float,
    n_bins: int = 60,
    n_bins_linear: int = 10,
) -> np.ndarray:
    """
    returns symmetrical logarithmic bins

    Bins have same absolute vmin, vmax if vmin is negative
    Parameters:
    ----------
    vmin: minimum value of data
    vmax: maximum value of data
    linear_thresh: threshold below which bins are linear to include zero values
    n_bins: number of bins for logarithmic part of bins
    n_bins_linear: number of bins for linear part of bins

    Returns
    -------
    symmetrical bin edges
    """
    if isinstance(vmin, np.datetime64) or vmin > 0:
        bins = 10 ** np.linspace(np.log10(vmin), np.log10(vmax), n_bins)
    elif vmin == 0:
        bins = np.hstack(
            (
                np.linspace(0, linear_thresh, n_bins_linear),
                10 ** np.linspace(np.log10(linear_thresh), np.log10(vmax)),
            )
        )
    else:
        bins = np.hstack(
            (
                -(
                    10
                    ** np.linspace(
                        np.log10(vmax),
                        np.log10(linear_thresh),
                        n_bins // 2,
                        endpoint=False,
                    )
                ),
                np.linspace(-linear_thresh, linear_thresh, n_bins_linear, endpoint=False),
                10 ** np.linspace(np.log10(linear_thresh), np.log10(vmax), n_bins // 2),
            )
        )
    return bins



import pathlib
import typing as t

import a2.plotting.axes_utils
import a2.plotting.utils_plotting
import matplotlib.pyplot as plt
import numpy as np

try:
    import sklearn.metrics
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(f"{e}\n" f"Optional dependencies missing, install via" " `poetry install --extras=train`")


def plot_prediction_certainty(
    truth: np.ndarray,
    prediction_probabilities: np.ndarray,
    n_true_labels: int = 2,
    n_bins: int = 10,
    label_x: str = "True label",
    label_y: str = "Prediction probability for 'raining'",
    label_colorbar: str = "Number of Tweets",
    filename: t.Union[str, pathlib.Path] = None,
    font_size: int = 12,
    vmin=None,
    vmax=None,
    cmap="viridis",
    figure_size: t.Sequence = None,
    return_matrix: bool = False,
    ax: plt.axes = None,
    fig: plt.figure = None,
    **kwargs,
) -> t.Union[plt.Axes, t.Tuple[plt.Axes, np.ndarray]]:
    """
    Compute 2d histogram of true labels (x-axis) and
    binned prediction probabilities (y-axis)

    Parameters:
    ----------
    truth: true labels
    prediction_probabilities: probabilities for labels
    n_true_labels: number of different labels
    n_bins: number of bins for prediction probabilities
    label_x: label of x-axis
    label_y: label of y-axis
    label_colorbar: label of colorbar axis
    filename: save figure to this file, not saved if `None`
    font_size: font size of all labels of plot
    figure_size: size of figure [width, height]
    vmin: Minimum value for colorbar
    vmax: Maximum value for colorbar
    return_matrix: return 2d histogram matrix and axes

    Returns
    -------
    axes of plot
    """
    if figure_size is None:
        figure_size = [8, 8]
    ax, H = a2.plotting.histograms.plot_histogram_2d(
        truth,
        prediction_probabilities,
        n_bins=[n_true_labels, n_bins],
        label_x=label_x,
        label_colorbar=label_colorbar,
        label_y=label_y,
        overplot_values=True,
        return_matrix=True,
        filename=filename,
        fig=fig,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        font_size=font_size,
        **kwargs,
    )
    if return_matrix:
        return ax, H
    return ax


def classification_report(
    truth: t.Sequence,
    prediction: t.Sequence,
    target_names: t.Optional[t.Sequence[str]] = None,
    output_dict: bool = True,
) -> t.Union[str, t.Mapping]:
    """
    Compute classification report, returns precision, recall, f1-score
    and respective averages

    report returned as dict or string (see `output_dict`)
    Parameters:
    ----------
    truth: true labels
    prediction: predicted labels
    target_names: names of labels
    output_dict: return report as dictionary rather than string

    Returns
    -------
    report
    """
    if target_names is None:
        target_names = ["not raining", "raining"]
    report = sklearn.metrics.classification_report(
        truth,
        prediction,
        target_names=target_names,
        output_dict=output_dict,
    )
    return report


def check_prediction(
    truth: t.Sequence,
    prediction: t.Sequence,
    output_dict: bool = True,
    filename: t.Union[str, pathlib.Path] | None = None,
):
    plot_confusion_matrix(
        truth,
        prediction,
        filename=filename,
        overplot_round_base=2,
    )

    report = classification_report(
        truth,
        prediction,
        target_names=["not raining", "raining"],
        output_dict=output_dict,
    )
    return report


def plot_confusion_matrix(
    truth: t.Sequence,
    prediction: t.Sequence,
    filename: t.Union[str, pathlib.Path] | None = None,
    normalize: str = "all",
    font_size: int = 14,
    figure_size: t.Sequence | None = None,
    label_x: str | None = "Predicted label",
    label_y: str | None = "True label",
    vmin: float | None = None,
    vmax: float | None = None,
    text_color: str | None = "firebrick",
    colormap: str | None = "Blues",
    overplot_round_base: int | None = None,
    tick_labels_x: tuple | None = ("not raining", "raining"),
    tick_labels_y: tuple | None = ("raining", "not raining"),
    colorbar_label: str | None = "Fraction",
    colorbar_width: float = 0.05,
    ax: object | None = None,
    ax_colorbar: object | None = None,
):
    """
    Computes classification report and plots confusion matrix

    Parameters:
    ----------
    truth: True labels
    prediction: Predicted labels
    filename: Save figure to this file, not saved if `None`
    normalize: Normalize style of confusion matrix
    output_dict: Return classification report as dictionary
    font_size: Font size of all labels of plot
    figure_size: Size of figure [width, height], default [6, 6]

    Returns
    -------
    classification report
    """
    if figure_size is None:
        figure_size = [6, 6]
    cm = sklearn.metrics.confusion_matrix(truth, prediction, normalize=normalize)
    print(f"{cm=}")
    if ax is None:
        fig, ax, ax_colorbar = a2.plotting.utils_plotting.create_axes_grid(
            1, 1, figure_size=figure_size, unravel=True, colorbar_off=False, colorbar_width=colorbar_width
        )
    else:
        fig = ax.get_figure()
    norm = a2.plotting.utils_plotting.get_norm("linear", vmin=vmin, vmax=vmax)
    xedges = [0, 0.5, 1]
    yedges = [0, 0.5, 1]
    cm = cm[::-1, :].T
    mesh = ax.pcolormesh(xedges, yedges, cm.T, norm=norm, cmap=colormap)
    cbar = plt.colorbar(mesh, cax=ax_colorbar, orientation="vertical")
    a2.plotting.axes_utils.set_colorbar(cbar.ax, label_y=colorbar_label, fontsize=font_size)
    a2.plotting.utils_plotting.overplot_values(
        np.array(cm), ax, 2, 2, color=text_color, round_to_base=overplot_round_base, font_size=font_size
    )
    a2.plotting.utils_plotting.set_axis_tick_labels(ax, [0.25, 0.75], tick_labels_x, "x")
    a2.plotting.utils_plotting.set_axis_tick_labels(ax, [0.25, 0.75], tick_labels_y, "y")
    a2.plotting.axes_utils.set_axes(
        ax,
        xlim=[0, 1],
        ylim=[0, 1],
        fontsize=font_size,
        label_x=label_x,
        label_y=label_y,
    )

    a2.plotting.utils_plotting.save_figure(fig, filename)
    return fig


def plot_roc(
    truth: t.Sequence,
    prediction_probabilities: t.Sequence,
    filename: t.Union[str, pathlib.Path] = None,
    font_size: int = 12,
    figure_size: t.Sequence = None,
    ax: object | None = None,
    fig: object | None = None,
    label_x: str | None = "False Positive Rate",
    label_y: str | None = "True Positive Rate",
    return_rates: bool = False,
) -> t.Union[plt.Axes, t.Tuple[plt.Axes, t.Sequence, t.Sequence]]:
    """
    Plots Receiver Operating Characteristic curve

    Can optionally return true positive and false positive rate
    (if `return_rates=True`)
    Parameters:
    ----------
    truth: true labels
    prediction_probabilities: probabilities for labels
    filename: save figure to this file, not saved if `None`
    font_size: font size of all labels of plot
    figure_size: size of figure [width, height], default: [6, 6]
    return_rates: return true positive and false positive rate

    Returns
    -------
    axes of plot
    """
    if figure_size is None:
        figure_size = [6, 6]
    false_positive_rate, true_positive_rate, _ = sklearn.metrics.roc_curve(
        truth, prediction_probabilities, drop_intermediate=False
    )
    roc_auc = sklearn.metrics.auc(false_positive_rate, true_positive_rate)
    fig, ax = a2.plotting.utils_plotting.create_figure_axes(
        fig=fig, ax=ax, figure_size=figure_size, font_size=font_size
    )
    lw = 2
    ax.plot(
        false_positive_rate,
        true_positive_rate,
        lw=lw,
        label="ROC curve (area = %.04f)" % roc_auc,
    )

    ax.plot([0, 1], [0, 1], color="navy", lw=lw, linestyle="--")
    a2.plotting.axes_utils.set_axes(
        ax,
        xlim=[0, 1],
        ylim=[0, 1],
        fontsize=font_size,
        label_x=label_x,
        label_y=label_y,
    )

    ax.legend(loc="lower right")
    a2.plotting.utils_plotting.save_figure(fig, filename)
    if return_rates:
        return ax, true_positive_rate, false_positive_rate
    return ax



import logging
import typing as t
import warnings

import a2.plotting
import mantik
import mlflow


def initialize_mantik():
    """
    Attemps to initialize mantik, throws exception if fails
    """
    try:
        mantik.init_tracking()
    except Exception as e:
        warnings.warn(f"{e}\nCannot initialize mantik!")


def log_metric_classification_report(truth: t.Sequence, predictions: t.Sequence, step: int = 1):
    """
    Compute f1 score and logs results to mlflow

    Parameters:
    ----------
    truth: True labels
    predictions: Predicted labels
    prediction_probabilities: Prediction probability for both labels, shape = [n_tests, 2]
    Step: Current training stop (epoch)

    Returns
    -------
    """
    initialize_mantik()
    classification_report = a2.plotting.analysis.check_prediction(
        truth=truth,
        prediction=predictions,
        filename="confusion_matrix.pdf",
        output_dict=True,
    )
    logging.info(classification_report)
    log_classification_report(classification_report, step)
    mlflow.log_artifact("confusion_matrix.pdf")


def log_classification_report(classification_report, step):
    initialize_mantik()
    mlflow.log_metric(
        key="eval_f1_raining",
        value=classification_report["raining"]["f1-score"],
        step=step,
    )
    mlflow.log_metric(
        key="eval_f1_not_raining",
        value=classification_report["not raining"]["f1-score"],
        step=step,
    )
    mlflow.log_metric(
        key="weighted average f1-score",
        value=classification_report["weighted avg"]["f1-score"],
        step=step,
    )
    mlflow.log_metric(
        key="macro average f1-score",
        value=classification_report["macro avg"]["f1-score"],
        step=step,
    )



import numpy as np
import pandas as pd


def time_breakdown(time: np.datetime64):
    """Breakdown time in individual components, i.e. year, month, day, hour, minute"""
    time_pandas = pd.to_datetime(time)
    year = time_pandas.year
    month = time_pandas.month
    day = time_pandas.day
    hour = time_pandas.hour
    minute = time_pandas.minute
    return year, month, day, hour, minute



import json
import logging
import os
import pathlib
import time
import typing as t

import a2.dataset.utils_dataset
import a2.utils.file_handling
import numpy as np
import pandas as pd
import pyproj
import shapely.geometry
import tqdm
import tweepy
import xarray


def _get_tweepy_api(wait_on_rate_limit: bool = True) -> tweepy.API:
    """
    returns Twitter api object from tweepy

    requires environment variables TWITTER_CONSUMER_KEY and TWITTER_ACCESS_TOKEN
    to work
    Parameters:
    ----------
    wait_on_rate_limit: api waits when too many requests send until
                        resend possible

    Returns
    -------
    tweepy api object
    """
    twitter_consumer_key = os.getenv("TWITTER_CONSUMER_KEY")
    if twitter_consumer_key is None:
        raise ValueError("tweepy api token: TWITTER_CONSUMER_KEY not found" "as environment variable!")
    twitter_access_token = os.getenv("TWITTER_ACCESS_TOKEN")
    if twitter_access_token is None:
        raise ValueError("tweepy api token: TWITTER_ACCESS_TOKEN not found" "as environment variable!")

    auth = tweepy.OAuth1UserHandler(
        twitter_consumer_key,
        twitter_access_token,
    )
    api = tweepy.API(
        auth,
        wait_on_rate_limit=wait_on_rate_limit,
        parser=tweepy.parsers.JSONParser(),
    )
    return api


def _append_not_found(place_id: str, filename: t.Union[str, pathlib.Path]) -> None:
    """
    append place_id's to csv file
    """
    if not a2.utils.file_handling.file_exists_and_not_empty(filename):
        logging.info(f"creating new csv file: {filename}")
        a2.utils.file_handling.csv_create(
            filename,
            ["# list of place_id's that were not found by twitter api\n"],
            [place_id],
            check_is_file_or_empty=True,
        )
    else:
        a2.utils.file_handling.csv_append(filename, [place_id])


def save_location_to_file(
    place_ids: t.Union[t.Sequence[str], set[str]],
    filename_location: t.Union[str, pathlib.Path] = "./locations.json",
    filename_location_not_found: t.Union[str, pathlib.Path] = "./locations_not_found.csv",
) -> None:
    """
    saves and queries json responses of information on place_id's to file

    Parameters:
    ----------
    place_ids: place_id's to query
    filename_location: name of file where json responses saved
    filename_location_not_found: adds place_id's to this csv file
                                 which cannot be found by twitter api

    Returns
    -------
    """
    api = _get_tweepy_api(wait_on_rate_limit=True)
    queried = 0

    # remove place_id's that have already been queried
    place_ids = set(place_ids)
    if os.path.isfile(filename_location):
        df = pd.read_json(filename_location)
        place_ids = set(place_ids) - set(df["id"].values)
        del df
    if os.path.isfile(filename_location_not_found):
        df = pd.read_csv(
            filename_location_not_found,
            skiprows=1,
            skip_blank_lines=False,
            names=["id"],
        )
        place_ids = set(place_ids) - set(df["id"].values)
        del df
    if len(place_ids):
        logging.info(f"Need to query {len(place_ids)} place_id's ...")

    for p_id in tqdm.tqdm(place_ids):
        if queried >= 74:
            time.sleep(60 * 15 + 15)
            queried = 0
        if os.path.isfile(filename_location):
            with open(filename_location) as f:
                data = json.load(f)
        queried += 1
        try:
            location = api.geo_id(place_id=p_id)
            print(f"{location=}")
        except Exception as e:
            _append_not_found(p_id, filename_location_not_found)
            logging.info(f"couldn't query: {p_id}")
            logging.info(e)
            continue
        if not os.path.isfile(filename_location):
            data = [location]
        else:
            data.append(location)
        a2.utils.file_handling.json_dump(filename_location, data, log_if_new_file=True)


def convert_coordinates_to_lat_long(
    ds: xarray.Dataset,
    key_coordinates: str = "geo.coordinates.coordinates",
    prefix_lat_long: str = "",
    overwrite: bool = False,
):
    """
    converts coordinates in dataset to latitude and longitude
    and adds them as new columns

    Parameters:
    ----------
    ds: dataset with coordinates specified
    key_coordinates: key name of coordinates (in string format)
    prefix_lat_long: prefix before latitude and longitude column names
    overwrite: allow overwriting of latitude and longitude column

    Returns
    -------
    dataset with longitude/latitude
    """
    if not overwrite and (f"{prefix_lat_long}longitude" in ds or f"{prefix_lat_long}latitude" in ds):
        raise RuntimeError("Fields latitude and/or longitude already present!")
    if key_coordinates not in ds:
        raise KeyError(f"coordinate_name: {key_coordinates} not found in dataframe!")

    if f"{prefix_lat_long}longitude" not in ds or overwrite:
        ds = a2.dataset.utils_dataset.initialize_variable(ds, f"{prefix_lat_long}longitude", ["index"], values=np.nan)
    if f"{prefix_lat_long}latitude" not in ds or overwrite:
        ds = a2.dataset.utils_dataset.initialize_variable(ds, f"{prefix_lat_long}latitude", ["index"], values=np.nan)
    if not a2.dataset.utils_dataset.is_same_type_data_array(ds, key_coordinates, which_type=str):
        ds[key_coordinates] = (
            ["index"],
            a2.dataset.utils_dataset.array_elements_to_str(ds[key_coordinates].values),
        )
    if a2.dataset.utils_dataset.is_same_type_data_array(ds, key_coordinates, which_type=str):
        s = ds[key_coordinates].str.extract(r"^\[([0-9.e-]+) ?,", dim=None)
        mask = s == ""
        s[mask] = np.nan
        ds[f"{prefix_lat_long}longitude"] = s.astype(float)
        s = ds[key_coordinates].str.extract(r" ?, ?([0-9.e-]+)\]$", dim=None)
        mask = s == ""
        s[mask] = np.nan
        ds[f"{prefix_lat_long}latitude"] = s.astype(float)

    else:
        raise Exception(f"do not understand datatype of field {key_coordinates} to convert!")

    return ds


def add_locations(
    ds: xarray.Dataset,
    filename_location: t.Union[str, pathlib.Path] = "./locations.json",
    filename_location_not_found: t.Union[str, pathlib.Path] = "./locations_not_found.csv",
    download: bool = True,
    key_place_id: str = "geo.place_id",
    key_coordinates: str = "geo.coordinates.coordinates",
) -> xarray.Dataset:
    """
    Adds location information to dataset based on place_id

    Information on places queried from twitter api/loaded from file if
    available. Add fields to dataset describing place_type,
    bounding_box, full_name and marks tweets if info was added in field
    coordinates_estimated. Centroid of inner boundary box is used as
    coordinate of location.
    Note, place_id's can correspond to arbitrary large entities.
    Parameters:
    ----------
    ds: dataset with place_id's specified
    filename_location: file path to json info on locations
    filename_location_not_found: file path to csv file with place_id's
                                 that couldn't be found in twitter query

    Returns
    -------
    dataset with location information
    """
    if download:
        save_location_to_file(
            list(ds[key_place_id].values),
            filename_location=filename_location,
            filename_location_not_found=filename_location_not_found,
        )

    ds_loc = pd.read_json(filename_location).to_xarray()
    ds = a2.dataset.utils_dataset.initialize_variable(ds, "coordinates_estimated", values=0, dtype=bool)
    ds = a2.dataset.utils_dataset.initialize_variable(ds, "centroid", dtype=object)
    ds = a2.dataset.utils_dataset.initialize_variable(ds, "place_type", dtype=object)
    ds = a2.dataset.utils_dataset.initialize_variable(ds, "bounding_box", dtype=object)
    ds = a2.dataset.utils_dataset.initialize_variable(ds, "full_name", dtype=object)

    tagged_found = 0
    places_in_dataset = set(ds[key_place_id].values)
    places_in_file = set(ds_loc.id.values)

    place_not_found = places_in_dataset - places_in_file
    places_in_file_array = np.array(list(places_in_file))
    places_in_dataset_array = np.array(list(places_in_dataset))
    logging.info(f"{len(place_not_found)} places not found in file, skipping those...")
    places_to_check = places_in_file_array[np.in1d(places_in_file_array, places_in_dataset_array)]

    for place_id in tqdm.tqdm(places_to_check):
        mask = np.logical_and(
            ds[key_place_id] == place_id,
            a2.dataset.utils_dataset.is_nan(ds, key_coordinates, dims=("index",)),
        )
        # only add location to places that do not already have autotracking
        # enabled, should be redundant however
        tagged_found += np.sum(mask.values)
        mask_loc = ds_loc.id == place_id
        if np.sum(mask_loc.values) != 1:
            raise ValueError(
                f"place_id {place_id} found multiple/zero times in location"
                "json! mask_loc.values: {mask_loc.values}, "
                f"np.sum(mask_loc.values): {np.sum(mask_loc.values)}"
            )
        ds["centroid"].loc[mask] = str(ds_loc["centroid"].loc[mask_loc].values[0])
        ds["place_type"].loc[mask] = ds_loc["place_type"].loc[mask_loc].values[0]
        ds["bounding_box"].loc[mask] = ds_loc["bounding_box"].loc[mask_loc].values[0]
        ds["full_name"].loc[mask] = ds_loc["full_name"].loc[mask_loc].values[0]
        ds["coordinates_estimated"].loc[mask] = True
    logging.info(f"found locations for {tagged_found} tweets")
    return ds


def _compute_bounding_box_area(bounding_box):
    """
    see compute_area_bounding_box
    """
    if hasattr(bounding_box, "item") and callable(getattr(bounding_box, "item")):
        bounding_box = bounding_box.item()
    try:
        if bounding_box == " " or pd.isnull(bounding_box):
            return -1
        else:
            lon, lat = zip(*bounding_box["coordinates"][0])
            pa = pyproj.Proj(f"+proj=aea +lat_1={lat[0]} +lat_2={lat[2]}" f" +lon_0={lon[0] + (lon[1] - lon[0]) / 2.}")
            x, y = pa(lon, lat)
            cop = {"type": "Polygon", "coordinates": [zip(x, y)]}
            return shapely.geometry.shape(cop).area / 1e6
    except Exception as e:
        raise Exception("Couldn't compute bounding box area of" f"bounding_box: {bounding_box}") from e


def compute_area_bounding_box(bounding_boxes: np.ndarray) -> list:
    """
    computes area of bounding box in km^2 (parallelized)

    an equal area projection is used to convert the coordinates
    (https://proj.org/operations/projections/aea.html)
    and then the area covered by the bounding box is computed.
    if no bounding box is given returns 0.
    if bounding box declaration is incorrect returns -1.
    Assumes that bounding boxes are provided in the form of dictionaries.
    Parameters:
    ----------
    bounding_boxes: array with bounding boxes

    Returns
    -------
    array with area of corresponding bounding box
    """
    bounding_box_area = a2.utils.utils.parallelize(_compute_bounding_box_area, bounding_boxes, single_arg=True)
    return bounding_box_area



import collections
import functools
import pathlib
import typing as t
import warnings

import a2.dataset.radar
import a2.dataset.tweets
import a2.plotting.axes_utils
import a2.plotting.utils_plotting
import a2.utils.checks
import a2.utils.constants
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray

TypeKeyTweets = object
TypeKeyRadar = object


def get_info_string_tweets(
    ds: xarray.Dataset,
    fields: t.List[str] = None,
    add: t.Optional[t.List[str]] = None,
) -> str:
    """
    Returns string giving information on fields +
    add keys of all Tweets in dataset

    Parameters:
    ----------
    ds: Dataset
    fields: Keys to print, default: "text", "latitude_rounded",
            "longitude_rounded", "created_at", "tp", "raining", "source"
    add: Additional keys to print

    Returns
    -------
    Info string
    """

    def get_present_variables(ds, fields):
        return [ds[f].values for f in fields if f in ds]

    if fields is None:
        fields = [
            "text",
            "latitude_rounded",
            "longitude_rounded",
            "created_at",
            "tp",
            "raining",
            "source",
        ]
    if add is not None:
        fields = fields + add
    fields = [f for f in fields if f in ds]
    to_print = ""
    if ds.index.shape:
        for tweet in zip(*get_present_variables(ds, fields)):
            for label, value in zip(fields, tweet):
                to_print += f"{label}: {value}\n"
    else:
        for label, value in zip(fields, get_present_variables(ds, fields)):
            to_print += f"{label}: {value}\n"
    return to_print


def plot_precipiation_map(
    ds_precipitation: xarray.Dataset,
    ds_tweets: xarray.Dataset,
    key_time: str = "created_at_h",
    key_longitude: str = "longitude_rounded",
    key_latitude: str = "latitude_rounded",
    key_tp: str = "tp",
    n_time: int = 1,
    delta_time: int = 1,
    delta_time_units: str = "h",
    delta_longitude: float = 1,
    delta_latitude: float = 1,
    filename: t.Union[str, pathlib.Path] = None,
    print_additional: t.Optional[t.List[str]] = None,
    add_time_before_plot: t.Optional[pd.Timedelta] = None,
    return_plots: bool = False,
) -> t.Union[plt.axes, t.Tuple[plt.axes, t.List]]:
    """
    Plot precipitation map corresponding to location and time of
    given Tweets and `n_time` plots in increments of `delta_time`
    [`delta_time_units`]

    Parameters:
    ----------
    ds_precipitation: Dataset of precipitation data
    ds_tweets: Dataset of Tweets
    key_time: Key of time variable
    key_longitude: Key of longitude variable
    key_latitude: Key of latitude variable
    key_tp: Key of latitude variable
    n_time: Number of additional delta times shown
            (1x +delta time, 1x-delta time -> n_time=1)
    delta_time: Value of time increments for additional plots
    delta_time_units: Units of time increments
    delta_longitude: Increment to include in weather maps to east and west
    delta_latitude: Increment to include in weather maps to north and south
    filename: Filename to save plot
    print_additional: Show additional fields of Tweet dataset
    add_time_before_plot: Increment time values from Tweet before looking
                          for corresponding value in precipitation dataset
    return_plots: Returns plot objects

    Returns
    -------
    axes(, plot object)
    """
    if print_additional is None:
        print_additional = []
    n_rows = ds_tweets.index.shape[0]
    n_cols = 1 + 2 * n_time
    fig, axes = plt.subplots(ncols=n_cols, nrows=n_rows, figsize=(15, 5 * n_rows))
    plots = np.full(np.shape(axes), np.nan, dtype=object)
    for i_tweet, (longitude, latitude, time) in enumerate(
        zip(
            ds_tweets[key_longitude].values,
            ds_tweets[key_latitude].values,
            ds_tweets[key_time].values,
        )
    ):
        for i_time, dt in enumerate(np.arange(-1 * delta_time * n_time, delta_time * (n_time + 1), delta_time)):
            ax = axes[i_tweet, i_time]
            time_to_plot = pd.to_datetime(time) + pd.Timedelta(f"{dt}{delta_time_units}")
            if add_time_before_plot:
                time_to_plot = time_to_plot + add_time_before_plot
            pcolormesh = (
                ds_precipitation[key_tp]
                .loc[f"{time_to_plot}"]
                .plot(
                    vmin=1e-6,
                    vmax=1e-3,
                    xlim=[
                        longitude - delta_longitude,
                        longitude + delta_longitude,
                    ],
                    ylim=[latitude - delta_latitude, latitude + delta_latitude],
                    ax=ax,
                    cmap="magma_r",
                )
            )
            plots[i_tweet, i_time] = pcolormesh
            ax.set_xlabel("")
            ax.set_ylabel("")
    for index, _ in enumerate(ds_tweets.index.values):
        title = get_info_string_tweets(
            a2.dataset.load_dataset.reset_index_coordinate(ds_tweets).sel(index=index),
            add=["preds_raining"] + print_additional,
        )
        axes[index, n_time].set_title(title)
    fig.tight_layout()
    a2.plotting.utils_plotting.save_figure(fig, filename)
    if return_plots:
        return axes, plots
    return axes


def circle_scatter(
    ax: plt.axes,
    x: np.ndarray,
    y: np.ndarray,
    color: t.Union[str, list, np.ndarray] = None,
    radius: t.Union[float, list, np.ndarray] = 1,
    fill: bool = True,
    cmap: matplotlib.colormaps = None,
    norm=None,
    alpha: float = 1,
    show_colorbar: bool = False,
):
    """
    Scatter plot with markers as circles, which allows for color mapping to colorbar.

    Parameters:
    ----------
    ax: matplotlib axes
    x: x coordinates on ax
    y: y coordinates on ax
    color: Color of circles, provide normed values if using colormap as color
    radius: Radius of circles in units of the plot
    fill: Wether to fille the circles with `color`
    cmap: Colormap used as color
    norm: Norm for colormap
    alpha: Transparency of circles
    show_colorbar: Show colorbar if using `cmap`

    Returns
    -------
    """

    if isinstance(radius, float) or isinstance(radius, int):
        radius = [radius for i in range(len(x))]
    if isinstance(color, str) or not isinstance(color, collections.abc.Iterable):
        color = [color for i in range(len(x))]
    if not fill:
        circles = [plt.Circle((xi, yi), color=c, radius=r, fill=fill) for xi, yi, r, c in zip(x, y, radius, color)]
        for c in circles:
            ax.add_artist(c)
    else:
        circles = [plt.Circle((xi, yi), radius=r) for xi, yi, r in zip(x, y, radius)]
        p = matplotlib.collections.PatchCollection(circles, cmap=cmap, norm=norm, alpha=alpha)
        p.set_array(color)
        ax.add_collection(p)
        cbar = None
        if cmap is not None and show_colorbar:
            cbar = plt.colorbar(p, ax=ax)
        return cbar


def _plot_radar_map_with_tweets(
    index: np.datetime64,
    ds: xarray.Dataset,
    path_to_dapceda: t.Optional[pathlib.Path] = None,
    key_tweets: TypeKeyTweets = None,
    key_radar: TypeKeyRadar = None,
    cumulative_radar: bool = False,
    cumulative_delta_time: float = 1,
    cumulative_delta_time_units: str = "h",
    vmin: t.Optional[float] = None,
    vmax: t.Optional[float] = None,
    xlim: t.Optional[list] = None,
    ylim: t.Optional[list] = None,
    fontsize: int = 20,
    marker_sizer_rescale: float = 0.01,
    colormap: str = "viridis",
    circle_alpha: float = 0.4,
    ax: t.Optional[plt.axes] = None,
    fig: t.Optional[plt.figure] = None,
    circle_size_constant: bool = False,
    circle_size: float = 0.1,
):
    """
    Plot background radar map with precipitation estimates for Tweets
    based on data from weather stations

    Parameters:
    ----------
    index: index in form of time
    ds: Tweet dataset that will be plotted
    path_to_dapceda: Path to folder with raw radar files
    key_tweets: Data class describing keys in dataframe for Tweets
    key_radar: Data class describing keys in dataframe for radar
    cumulative_radar: Radar is computed cumulatively
    cumulative_delta_time: Time range for which rada data is summed cumulatively
    cumulative_delta_time_units: Units of `cumulative_delta_time`
    vmin: Minimum value of colorscale depicting precipitation
    vmax: Maximum value of colorscale depicting precipitation
    xlim: Limits of x-axis showing longitude
    ylim: Limits of y-axis showing latitude
    fontsize: Font size of all text in plot
    marker_sizer_rescale: Rescaling factor for marker scaling (default 0.01: [km] * 0.01 ~ [deg lat/lon])
    colormap: Colormap used for precipitation
    circle_alpha: Alpha value for circle plot
    ax: Mapltotlib axes for plot
    fig: Figure of plot
    circle_size_constant: Setting circle size to constant value as specified here

    Returns
    -------
    """
    path_to_dapceda = a2.utils.file_handling.get_folder_radar(path_to_dapceda)
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    key_radar = a2.dataset.radar.get_key_radar(key_radar, cumulative_radar=cumulative_radar)
    time = index
    if ax is None:
        fig, ax = a2.plotting.utils_plotting.create_figure_axes(aspect="equal")
    if xlim is None:
        xlim = a2.utils.constants.UK_LONGITUDE_LIMIT
    if ylim is None:
        ylim = a2.utils.constants.UK_LATITUDE_LIMIT
    cmap = matplotlib.colormaps[colormap]
    if cumulative_radar:
        ds_radar = a2.dataset.radar.nimrod_ds_cumulative_from_time(
            path_to_dapceda, time, time_delta=cumulative_delta_time, time_delta_units=cumulative_delta_time_units
        )
    else:
        ds_radar = a2.dataset.radar.nimrod_ds_from_time(path_to_dapceda, time)
    if ds_radar is None:
        return
    tp = ds_radar[key_radar.tp].values
    if vmin is None:
        vmin = np.nanmin(tp)
    if vmax is None:
        vmax = np.nanmax(tp)
    norm = a2.plotting.utils_plotting.get_norm(None, vmin=vmin, vmax=vmax)
    pc = ds_radar[key_radar.tp][0].plot(
        x="longitude", y="latitude", ax=ax, xlim=xlim, ylim=ylim, norm=norm, cmap=colormap
    )
    a2.plotting.axes_utils.set_colorbar(pc.colorbar.ax, f"{key_radar.tp}", fontsize)

    size_marker = _retrieve_size_marker(
        ds, circle_size_constant, key_tweets.distance_stations, marker_sizer_rescale, circle_size
    )
    color_normed = [norm(x) for x in ds[key_tweets.station_tp].values]
    circle_scatter(
        ax,
        *(ds.longitude.values, ds.latitude.values),
        radius=size_marker,
        color=color_normed,
        cmap=cmap,
        norm=norm,
        alpha=circle_alpha,
        show_colorbar=False,
    )
    circle_scatter(ax, *(ds.longitude.values, ds.latitude.values), radius=size_marker, color="red", fill=False)
    _set_axes_weather_map(ax, xlim, ylim, fontsize, f"{pd.to_datetime(time)}")
    ax.set_aspect("equal")


def _retrieve_size_marker(
    ds: xarray.Dataset,
    circle_size_constant: bool,
    key_twitter_distance_station: str,
    marker_sizer_rescale: float,
    circle_size: float = 0.1,
) -> np.ndarray:
    """Get the size of the markers (e.g, circles) in correct shape"""
    if circle_size_constant or key_twitter_distance_station not in ds:
        size_marker = np.array([circle_size for _ in ds.latitude.values])
    else:
        size_marker = ds[key_twitter_distance_station].values * marker_sizer_rescale
    return size_marker


def _plot_tp_station_tweets(
    index,
    ds: xarray.Dataset,
    df_stations: pd.DataFrame,
    key_tweets: TypeKeyTweets = None,
    key_station_tp: str = "prcp_amt",
    key_station_time: str = "ob_end_time",
    vmin: float = None,
    vmax: float = None,
    xlim: t.Sequence = None,
    ylim: t.Sequence = None,
    time_offset_station: float = 0.5,
    time_offset_station_units: str = "h",
    fontsize: float = 20,
    marker_sizer_rescale: float = 0.01,
    colormap: str = "viridis",
    circle_alpha: float = 0.8,
    circle_size_constant: bool = False,
    circle_size: float = 0.1,
    radius_station: float = 0.1,
    ax: plt.axes = None,
    fig: plt.figure = None,
):
    """
    Plot station precipitation values and values assigned to Tweets

    Parameters:
    ----------
    index: Index in form of time
    ds: Tweet dataset that will be plotted
    df_stations: Data frame of weather station data
    key_tweets: Data class describing keys in dataframe for Tweets
    key_station_tp: Key in station dataframe for precipitation
    key_station_time: Key in station dataframe for time
    vmin: Minimum value of colorscale depicting precipitation
    vmax: Maximum value of colorscale depicting precipitation
    xlim: Limits of x-axis showing longitude
    ylim: Limits of y-axis showing latitude
    time_offset_station: Adding this offset to index (time) and get values of weather station at this time
    time_offset_station_units: Units of `time_offset_station`
    fontsize: Font size of all text in plot
    marker_sizer_rescale: Rescaling factor for marker scaling (default 0.01: [km] * 0.01 ~ [deg lat/lon])
    radius_station: Size of marker for station values
    circle_size_constant: Setting circle size to constant value as specified here
    colormap: Colormap used for precipitation
    circle_alpha: Alpha value for circle plot
    ax: Mapltotlib axes for plot
    fig: Figure of plot

    Returns
    -------
    """
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    time = index
    if ax is None:
        fig, ax = a2.plotting.utils_plotting.create_figure_axes(aspect="equal")
    if xlim is None:
        xlim = a2.utils.constants.UK_LONGITUDE_LIMIT
    if ylim is None:
        ylim = a2.utils.constants.UK_LATITUDE_LIMIT
    tp_twitter = ds[key_tweets.station_tp].values
    time_station = time + pd.Timedelta(f"{time_offset_station}{time_offset_station_units}")
    df_sel = df_stations[(df_stations[key_station_time] == pd.to_datetime(time_station))].drop_duplicates(
        keep="last", subset=["latitude", "longitude"]
    )
    tp_station = df_sel[key_station_tp].values
    if vmin is None:
        vmin = np.nanmin(tp_twitter)
    if vmax is None:
        vmax = np.nanmax(tp_twitter)
    cmap = matplotlib.colormaps[colormap]
    norm = a2.plotting.utils_plotting.get_norm(None, vmin=vmin, vmax=vmax)
    color_normed_tweets = [norm(x) for x in tp_twitter]
    size_marker = _retrieve_size_marker(
        ds, circle_size_constant, key_tweets.distance_stations, marker_sizer_rescale, circle_size
    )
    ax.text(0.1, 0.1, f"time_station: {time_station}", transform=ax.transAxes, fontsize=fontsize)
    colorbar = circle_scatter(
        ax,
        *(ds.longitude.values, ds.latitude.values),
        radius=size_marker,
        color=color_normed_tweets,
        cmap=cmap,
        norm=norm,
        alpha=circle_alpha,
        show_colorbar=True,
    )
    if colorbar is not None:
        a2.plotting.axes_utils.set_colorbar(colorbar.ax, "tp_mm", fontsize)

    color_normed_stations = [norm(x) for x in tp_station]
    circle_scatter(
        ax,
        *(df_sel.longitude.values, df_sel.latitude.values),  # type: ignore
        radius=radius_station,
        cmap=cmap,
        norm=norm,
        color=color_normed_stations,
        alpha=circle_alpha,
    )
    circle_scatter(
        ax,
        *(df_sel.longitude.values, df_sel.latitude.values),  # type: ignore
        radius=radius_station,
        color="red",
        fill=False,
    )
    circle_scatter(ax, *(ds.longitude.values, ds.latitude.values), radius=size_marker, color="blue", fill=False)

    _set_axes_weather_map(ax, xlim, ylim, fontsize, f"{pd.to_datetime(time)}")
    ax.set_aspect("equal")


def _set_axes_weather_map(ax, xlim, ylim, fontsize, title):
    """Predefine subset of axes properties required for all weather maps, based on `a2.plotting.axes_utils.set_axes`"""
    a2.plotting.axes_utils.set_axes(
        ax, xlim=xlim, ylim=ylim, fontsize=fontsize, title=title, label_x="longitude", label_y="latitude"
    )


class MapsPlotter:
    """
    Class used to plot weather maps with multiple axes.
    Determines index based on `choice_type` which indicates time shown in individual axes.
    Selects subset of dataset `ds` to be plotted on individual axes based on `selector`.
    """

    def __init__(
        self,
        ds: xarray.Dataset,
        grid_shape: t.Tuple[int, ...],
        selector: t.Callable,
        figsize_default: float = 15,
        figsize: t.Optional[t.List[float]] = None,
        key_time: str = None,
        choice_type: str = "random",
        increment_time_value: np.datetime64 = None,
        increment_time_delta: float = None,
        increment_time_delta_units: str = None,
        aspect: str = "auto",
        height_ratios: t.Optional[t.List] = None,
    ) -> None:
        """
        Initializes class including setting up selection and choice.

        Parameters:
        ----------
        ds: Dataset to be plotted
        grid_shape: Shape of axes grid (rows, columns)
        selector: Determines which Tweets to select for specific axes based on `choice_type`
        figsize_default: Default figsize along largest axis
        figsize: Define size of figure
        key_time: Choice is computed based on this key in `ds`
        choice_type: Choice for index of individual axes:
            "random": Returns random index [int] based on size of `ds`
            "random_time": Returns randomly selected time from `ds[key_time]`
            "increment_time": Returns times as increments from `increment_time_value` in steps
                              `increment_time_delta` [`increment_time_delta_units`]
        increment_time_value: Starting time for `choice_type`="increment_time"
        increment_time_delta: Delta time for `choice_type`="increment_time"
        increment_time_delta_units: Delta time units `choice_type`="increment_time"
        aspect: Aspect of axes ("auto"/"equal")
        height_ratios: Ratios of height of axes along y-axis of figure

        Returns
        -------
        """
        self.ds = ds
        self.grid_shape = grid_shape
        self.selector = selector
        self.figsize_default = figsize_default
        self.increment_time_value = increment_time_value
        self.increment_time_delta = increment_time_delta
        self.increment_time_delta_units = increment_time_delta_units
        self.figsize = self._set_figsize(figsize=figsize)
        self.key_time = key_time
        self.n_sample = self.grid_shape[0] * self.grid_shape[1]
        self.choices = self._set_choices(choice_type=choice_type)
        self.ds_selected = self._do_selection()
        self.aspect = aspect
        self.height_ratios = height_ratios

    def _validate_increment_time(self):
        a2.utils.checks.assert_type(self.increment_time_value, np.datetime64)
        a2.utils.checks.assert_type(self.increment_time_delta, float | int)
        a2.utils.checks.assert_type(self.increment_time_delta_units, str)

    def _set_figsize(self, figsize: t.Optional[t.List[float]]):
        if figsize is None:
            figsize = [0.0, 0.0]
            num_plots_max_axis = 0 if self.grid_shape[0] > self.grid_shape[1] else 1
            figsize[num_plots_max_axis] = self.figsize_default
            figsize[1 - num_plots_max_axis] = (
                self.figsize_default * self.grid_shape[1 - num_plots_max_axis] / self.grid_shape[num_plots_max_axis]
            )
            figsize = figsize[::-1]
        return figsize

    def _do_selection(self):
        ds_selected = []
        for choice in self.choices:
            ds_selected.append(self.selector(choice, self.ds))
        return ds_selected

    def _set_choices(self, choice_type: str) -> list:
        ds = self.ds
        if choice_type == "random":
            self.size_data = len(ds.index.values)
            return list(a2.utils.utils.get_random_indices(self.n_sample, self.size_data))
        elif choice_type == "random_time":
            time_values = ds[self.key_time].values
            self.size_data = len(time_values)
            return [time_values[i] for i in a2.utils.utils.get_random_indices(self.n_sample, self.size_data)]
        elif choice_type == "increment_time":
            self._validate_increment_time()
            start_time = pd.to_datetime(self.increment_time_value)  # type: ignore
            times = []
            if not isinstance(self.increment_time_delta, (int, float)):
                raise TypeError(f"{self.increment_time_delta=} should be float!")
            for i in range(self.n_sample):
                times.append(
                    start_time + pd.Timedelta(f"{i*self.increment_time_delta}{self.increment_time_delta_units}")
                )
            return times
        else:
            raise NotImplementedError(f"{choice_type=} not implemented!")

    def plot(self, plot_fn: t.Callable, processes: int = -1):
        """plot axes in parallel"""
        self.fig = a2.plotting.parallel_plotting.parallel_plot(
            plot_fn=plot_fn,
            data=zip(self.choices, self.ds_selected),
            grid_shape=self.grid_shape,
            figsize=self.figsize,
            aspect=self.aspect,
            height_ratios=self.height_ratios,
            processes=processes,
        )
        return self.fig


def plot_radar_map_with_tweets(
    ds: xarray.Dataset,
    grid_shape: t.Optional[t.Tuple[int, int]] = None,
    figsize: t.Optional[t.List] = None,
    figsize_default: float = 30,
    path_to_dapceda: t.Optional[pathlib.Path] = None,
    key_tweets: TypeKeyTweets = None,
    key_radar: TypeKeyRadar = None,
    vmin: t.Optional[float] = None,
    vmax: t.Optional[float] = None,
    xlim: t.Optional[list] = None,
    ylim: t.Optional[list] = None,
    choice_type: str = "random_time",
    choice_time: str = "time_radar",
    selection_delta_time: float = 10,
    selection_delta_time_units: str = "m",
    selection_key_twitter_time: str = "time_radar",
    selector_use_limits: t.Sequence = None,
    increment_time_value: np.datetime64 = None,
    increment_time_delta: float = 1,
    increment_time_delta_units: str = "h",
    colormap: str = "viridis",
    circle_alpha: float = 1,
    cumulative_radar: bool = False,
    cumulative_delta_time: float = 1,
    cumulative_delta_time_units: str = "h",
    circle_size_constant: bool = False,
    processes: int = -1,
) -> plt.figure:
    """
    Plot background radar map with precipitation estimates for Tweets
    based on data from weather stations

    Parameters:
    ----------
    ds: Tweet dataset that will be plotted
    grid_shape: Shape of axes grid (rows, columns), default: (2, 1)
    figsize: Define size of figure
    figsize_default: Default figsize along largest axis
    path_to_dapceda: Path to folder with raw radar files
    key_tweets: Data class describing keys in dataframe for Tweets
    key_radar: Data class describing keys in dataframe for radar
    vmin: Minimum value of colorscale depicting precipitation
    vmax: Maximum value of colorscale depicting precipitation
    xlim: Limits of x-axis showing longitude
    ylim: Limits of y-axis showing latitude
    choice_type: Choice for index of individual axes:
        "random": Returns random index [int] based on size of `ds`
        "random_time": Returns randomly selected time from `ds[key_time]`
        "increment_time": Returns times as increments from `increment_time_value`
                          in steps `increment_time_delta` [`increment_time_delta_units`]
    choice_time: Choice is computed based on this key in `ds`
    selection_delta_time: Delta time within which selection from dataset made
    selection_delta_time_units: Units for selection_delta_time
    selection_key_twitter_time: Selection from dataset for individual axes based on this time key
    selector_use_limits: Use range in time within -delta time up to +delta time,
                         include/exclude limits as [True/False, True/False]
    increment_time_value: Starting time for `choice_type`="increment_time"
    increment_time_delta: Delta time for `choice_type`="increment_time"
    increment_time_delta_units: Delta time units `choice_type`="increment_time"
    colormap: Colormap used for precipitation
    circle_alpha: Alpha value for circle plot
    cumulative_radar: Radar is computed cumulatively
    cumulative_delta_time: Time range for which rada data is summed cumulatively
    cumulative_delta_time_units: Units of `cumulative_delta_time`
    circle_size_constant: Setting circle size to constant value as specified here
    processes: Number of plots (axes) generated in parallel

    Returns
    -------
    """
    if grid_shape is None:
        grid_shape = (2, 1)
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    partial_time_selector = functools.partial(
        time_selector,
        delta_time=selection_delta_time,
        delta_time_units=selection_delta_time_units,
        key_time=selection_key_twitter_time,
        use_limits=selector_use_limits,
    )
    p = MapsPlotter(
        ds,
        grid_shape=grid_shape,
        figsize_default=figsize_default,
        figsize=figsize,
        choice_type=choice_type,
        key_time=choice_time,
        selector=partial_time_selector,
        increment_time_value=increment_time_value,
        increment_time_delta=increment_time_delta,
        increment_time_delta_units=increment_time_delta_units,
        aspect="equal",
    )
    partial_plot_radar_map_with_tweets = functools.partial(
        _plot_radar_map_with_tweets,
        path_to_dapceda=path_to_dapceda,
        key_tweets=key_tweets,
        key_radar=key_radar,
        cumulative_radar=cumulative_radar,
        cumulative_delta_time=cumulative_delta_time,
        cumulative_delta_time_units=cumulative_delta_time_units,
        vmin=vmin,
        vmax=vmax,
        xlim=xlim,
        ylim=ylim,
        colormap=colormap,
        circle_alpha=circle_alpha,
        circle_size_constant=circle_size_constant,
    )
    fig = p.plot(partial_plot_radar_map_with_tweets, processes=processes)
    return fig


def plot_tp_station_tweets(
    ds: xarray.Dataset,
    df_stations: pd.DataFrame,
    grid_shape: t.Optional[t.Tuple] = None,
    figsize: t.Optional[t.List] = None,
    figsize_default: float = 30,
    key_tweets: TypeKeyTweets = None,
    key_station_tp: str = "prcp_amt",
    key_station_time: str = "ob_end_time",
    vmin: float = None,
    vmax: float = None,
    xlim: t.Sequence = None,
    ylim: t.Sequence = None,
    time_offset_station: float = 0.5,
    time_offset_station_units: str = "h",
    selector_key_time: str = "created_at",
    selection_delta_time: float = 0.5,
    selection_delta_time_units: str = "h",
    selector_use_limits: t.Sequence = None,
    choice_type: str = "random_time",
    choice_time: str = "time_half",
    increment_time_value=None,
    increment_time_delta: float = 1,
    increment_time_delta_units: str = "h",
    fontsize: float = 20,
    marker_sizer_rescale: float = 0.01,
    colormap: str = "viridis",
    circle_alpha: float = 0.8,
    circle_size_constant: bool = False,
    circle_size: float = 0.1,
    processes: int = -1,
) -> plt.figure:
    """
    Plot station precipitation values and values assigned to Tweets

    Parameters:
    ----------
    ds: Tweet dataset that will be plotted
    df_stations: Data frame of weather station data
    grid_shape: Shape of axes grid (rows, columns)
    figsize: Define size of figure
    figsize_default: Default figsize along largest axis
    key_tweets: Data class describing keys in dataframe for Tweets
    key_station_tp: Key in station dataframe for precipitation
    key_station_time: Key in station dataframe for time
    vmin: Minimum value of colorscale depicting precipitation
    vmax: Maximum value of colorscale depicting precipitation
    xlim: Limits of x-axis showing longitude
    ylim: Limits of y-axis showing latitude
    time_offset_station: Adding this offset to index (time) and get values of weather station at this time
    time_offset_station_units: Units of `time_offset_station`
    selector_key_time: Selection from dataset for individual axes based on this time key
    selection_delta_time: Delta time within which selection from dataset made
    selection_delta_time_units: Units for selection_delta_time
    selector_use_limits: Use range in time within -delta time up to +delta time,
                         include/exclude limits as [True/False, True/False]
    choice_type: Choice for index of individual axes:
        "random": Returns random index [int] based on size of `ds`
        "random_time": Returns randomly selected time from `ds[key_time]`
        "increment_time": Returns times as increments from `increment_time_value`
                          in steps `increment_time_delta` [`increment_time_delta_units`]
    choice_time: Choice is computed based on this key in `ds`
    increment_time_value: Starting time for `choice_type`="increment_time"
    increment_time_delta: Delta time for `choice_type`="increment_time"
    increment_time_delta_units: Delta time units `choice_type`="increment_time"
    fontsize: Font size of all text in plot
    marker_sizer_rescale: Rescaling factor for marker scaling (default 0.01: [km] * 0.01 ~ [deg lat/lon])
    colormap: Colormap used for precipitation
    circle_alpha: Alpha value for circle plot
    circle_size_constant: Setting circle size to constant value as specified here
    processes: Number of plots (axes) generated in parallel

    Returns
    -------
    figure
    """
    if grid_shape is None:
        grid_shape = (2, 1)
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    partial_time_selector = functools.partial(
        time_selector,
        delta_time=selection_delta_time,
        delta_time_units=selection_delta_time_units,
        key_time=selector_key_time,
        use_limits=selector_use_limits,
    )
    p = MapsPlotter(
        ds,
        grid_shape=grid_shape,
        figsize_default=figsize_default,
        figsize=figsize,
        choice_type=choice_type,
        key_time=choice_time,
        selector=partial_time_selector,
        increment_time_value=increment_time_value,
        increment_time_delta=increment_time_delta,
        increment_time_delta_units=increment_time_delta_units,
        aspect="equal",
        height_ratios=None,
    )
    partial_plot_tp_station_tweets = functools.partial(
        _plot_tp_station_tweets,
        df_stations=df_stations,
        key_tweets=key_tweets,
        key_station_tp=key_station_tp,
        key_station_time=key_station_time,
        vmin=vmin,
        vmax=vmax,
        xlim=xlim,
        ylim=ylim,
        time_offset_station=time_offset_station,
        time_offset_station_units=time_offset_station_units,
        fontsize=fontsize,
        marker_sizer_rescale=marker_sizer_rescale,
        colormap=colormap,
        circle_size_constant=circle_size_constant,
        circle_size=circle_size,
        circle_alpha=circle_alpha,
    )
    fig = p.plot(partial_plot_tp_station_tweets, processes=processes)
    return fig


def time_selector(
    time,
    ds: xarray.Dataset,
    delta_time: float = 0.5,
    delta_time_units: str = "h",
    key_time: str = "created_at",
    use_limits: t.Optional[list] = None,
) -> xarray.Dataset:
    """
    Select part of dataset `ds` where `key_time` is within upper and/or lower delta time around `time`

    Parameters:
    ----------
    time: Time
    ds: Dataset to select from
    delta_time: Delta making up lower and/or upper limit of selected period of time
    delta_time_units: Units of `delta_time`
    key_time: Key in dataset that defines time
    use_limits: Use range in time within -delta time up to +delta time,
                include/exclude limits as [True/False, True/False]

    Returns
    -------
    dataset
    """
    if use_limits is None:
        use_limits = [True, True]
    time_lower_limit = pd.to_datetime(time) - pd.Timedelta(f"{delta_time}{delta_time_units}")
    time_upper_limit = pd.to_datetime(time) + pd.Timedelta(f"{delta_time}{delta_time_units}")
    if use_limits[0] and use_limits[1]:
        return ds.where((ds[key_time] >= time_lower_limit) & (ds[key_time] <= time_upper_limit), drop=True)
    elif use_limits[0]:
        return ds.where((ds[key_time] >= time_lower_limit) & (ds[key_time] <= time), drop=True)
    elif use_limits[1]:
        return ds.where((ds[key_time] >= time) & (ds[key_time] <= time_upper_limit), drop=True)
    else:
        warnings.warn(f"Time selector chosen but not limits {use_limits=} used: returning full dataset!")
        return ds


def plot_radar_from_time(
    time,
    path_to_dapceda: pathlib.Path = None,
    cumulative: bool = False,
    time_delta: float = 1,
    time_delta_units: str = "h",
    vmin: float = None,
    vmax: float = None,
    ax: plt.axes = None,
    xlim: t.Optional[list] = None,
    ylim: t.Optional[list] = None,
) -> xarray.Dataset:
    """
    Plot radar map at certain time.
    Returns radar dataset.

    Parameters:
    ----------
    time: Time
    path_to_dapceda: Path to folder with raw radar files
    cumulative: Radar is computed cumulatively
    time_delta: Time range for which rada data is summed cumulatively
    time_delta_units: Units of `time_delta`
    vmin: Minimum value of colorscale depicting precipitation
    vmax: Maximum value of colorscale depicting precipitation
    ax: Mapltotlib axes for plot
    xlim: Limits of x-axis showing longitude
    ylim: Limits of y-axis showing latitude

    Returns
    -------
    dataset
    """
    if path_to_dapceda is None:
        path_to_dapceda = a2.utils.file_handling.get_folder_radar(path_to_dapceda)
    if ax is None:
        fig, ax = a2.plotting.utils_plotting.create_figure_axes(aspect="equal")
    if xlim is None:
        xlim = a2.utils.constants.UK_LONGITUDE_LIMIT
    if ylim is None:
        ylim = a2.utils.constants.UK_LATITUDE_LIMIT
    if cumulative:
        ds = a2.dataset.radar.nimrod_ds_cumulative_from_time(
            path_to_dapceda, time, time_delta=time_delta, time_delta_units=time_delta_units
        )
        key_tp = "tp_mm_cum"
    else:
        ds = a2.dataset.radar.nimrod_ds_from_time(path_to_dapceda, time)
        key_tp = "tp_mm"

    ds[key_tp][0].plot(
        x="longitude",
        y="latitude",
        vmin=vmin,
        vmax=vmax,
        ax=ax,
        xlim=xlim,
        ylim=ylim,
    )

    return fig, ds



import pathlib

import matplotlib.axes
import matplotlib.figure
import pandas as pd


UK_LONGITUDE_LIMIT = [-9, 3]
UK_LATITUDE_LIMIT = [49, 61]
PATH_DAPCEDA = pathlib.Path("/home/kristian/Projects/a2/notebooks/dataset/dap.ceda.ac.uk")
TIME_TYPE_PANDAS = pd.Timestamp
TYPE_MATPLOTLIB_FIGURES = matplotlib.figure.Figure
TYPE_MATPLOTLIB_AXES = matplotlib.axes.Axes



import dataclasses
import functools
import logging
import typing as t

import a2.dataset.tweets
import a2.utils
import geopy.distance
import numpy as np
import pandas as pd
import xarray


@dataclasses.dataclass(frozen=True)
class KeyStations:
    """Dataclass storing key convention for weather station datasets"""

    latitude = "latitude"
    longitude = "longitude"
    time = "ob_end_time"
    time_shifted = "ob_end_time_shifted"
    precipitation = "prcp_amt"


def distance_between_coordinates_in_km(
    latitude_1: float, longitude_1: float, latitude_2: float, longitude_2: float
) -> float:
    """Computes the the distance between two pairs of long/lat"""
    return geopy.distance.geodesic((latitude_1, longitude_1), (latitude_2, longitude_2)).km


def _find_closest_coordinate_and_distance(
    tweet_latitude: float,
    tweet_longitude: float,
    weather_station_latitudes: np.ndarray,
    weather_station_longitudes: np.ndarray,
):
    """
    Find the closest weather station to a Tweet's long/lat,
    returns the weather stations coordinates and distance

    Parameters:
    ----------
    tweet_latitude: Latitude of the Tweet
    tweet_longitude: Longitude of the Tweet
    weather_station_latitudes: Latitudes of all weather stations
    weather_station_longitudes: Longitudes of all weather stations

    Returns
    (latitude, longitude) of closest weather station, distance to closest weather station
    -------
    """
    distances = np.array(
        [
            distance_between_coordinates_in_km(tweet_latitude, tweet_longitude, lat, long)
            for lat, long in zip(weather_station_latitudes, weather_station_longitudes)
        ]
    )
    index_min = np.argmin(distances)
    latitude = weather_station_latitudes[index_min]
    longitude = weather_station_longitudes[index_min]
    distance = distances[index_min]
    return (latitude, longitude), distance


def find_closest_coordinate_and_distance(
    tweet_latitudes: np.ndarray,
    tweet_longitudes: np.ndarray,
    weather_station_latitudes: np.ndarray,
    weather_station_longitudes: np.ndarray,
    processes: int = -1,
) -> t.Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Find the closest weather station to a list of long/lat pairs of Tweets,
    returns the weather stations coordinates and distance
    Parallelizes `_find_closest_coordinate_and_distance`

    Parameters:
    ----------
    tweet_latitudes: Latitudes of Tweets
    tweet_longitudes: Longitudes of Tweets
    weather_station_latitudes: Latitudes of all weather stations
    weather_station_longitudes: Longitudes of all weather stations
    weather_station_longitudes: Number of tasks run in parallel

    Returns
    latitudes, longitudes, distances of/to closest weather stations
    -------
    """
    function = functools.partial(
        _find_closest_coordinate_and_distance,
        weather_station_latitudes=weather_station_latitudes,
        weather_station_longitudes=weather_station_longitudes,
    )
    results = a2.utils.utils.parallelize(
        function, args_zipped=zip(tweet_latitudes, tweet_longitudes), processes=processes
    )
    latitudes = []
    longitudes = []
    distances = []
    for x in results:
        latitude_longitude, dist = x
        latitudes.append(latitude_longitude[0])
        longitudes.append(latitude_longitude[1])
        distances.append(dist)
    return np.array(latitudes), np.array(longitudes), np.array(distances)


def unique_coordinates(df: pd.DataFrame, coordinates: list = None):
    """Finds unique coordinate pairs"""
    if coordinates is None:
        coordinates = ["latitude", "longitude"]
    return np.unique(np.array([df[x].values for x in coordinates]), axis=1)


def add_station_precipitation(
    ds: xarray.Dataset,
    df_stations: pd.DataFrame,
    key_tweets=None,
    key_stations=None,
    station_time_offset: str = "30m",
    coords: t.Optional[list] = None,
    processes: int = -1,
):
    """
    Add precipitation values from nearest weather station to Tweet's location

    Parameters:
    ----------
    ds: Dataset of Tweets
    df_stations: Dataframe with weather station data
    key_tweets: Dataclass storing key convention for tweets datasets
    key_stations: Dataclass storing key convention for weather station datasets
    station_time_offset: Time stamp of weather station measurement reduced by this amount
    coords: Coordinates of ds
    processes: Number of closest station found in parallel

    Returns
    ds with added weather station precipitation
    -------
    """
    if key_stations is None:
        key_stations = KeyStations()
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    df_stations = df_stations.copy()
    if coords is None:
        coords = list(ds.coords.keys())
    time_offset, unit_offset = a2.utils.utils.str_to_delta_time(station_time_offset)
    df_stations[key_stations.time_shifted] = pd.to_datetime(df_stations[key_stations.time]) - np.timedelta64(
        int(time_offset), unit_offset
    )  # type: ignore

    _assign_coordinates_of_nearest_station_to_tweets(ds, df_stations, key_tweets, coords, processes)

    ds[key_tweets.station_tp] = (
        coords,
        np.full_like(ds[a2.dataset.utils_dataset.get_variable_name_first(ds)].values, np.nan, dtype=float),
    )
    n_time_not_found_total = 0

    for latitude, longitude in zip(*unique_coordinates(df_stations)):
        mask_ds_station_location = (ds[key_tweets.latitude_stations] == latitude) & (
            ds[key_tweets.longitude_stations] == longitude
        )
        df_single_station = _get_station_from_lat_long(df_stations, latitude, longitude)
        times_of_tweets = ds[key_tweets.time_stations].loc[mask_ds_station_location].values
        if np.shape(times_of_tweets)[0] == 0:
            continue
        n_time_not_found, n_times_stations = _initialize_debug_counters()
        for time in times_of_tweets:
            n_times_stations += 1
            tp_match = df_single_station[key_stations.precipitation][
                df_single_station[key_stations.time_shifted] == time
            ]
            if tp_match.shape[0] > 1:
                logging.warning(
                    f"Found more than one entry ({tp_match}) for rain at weather station "
                    f"at time {time}: {tp_match=} with times "
                    f"{df_single_station[key_stations.time_shifted][df_single_station[key_stations.time_shifted] == time]}"  # noqa
                )
            elif tp_match.shape[0] == 0:
                _increment_debug_counters(n_time_not_found_total, n_time_not_found)
                continue
            ds[key_tweets.station_tp].loc[
                mask_ds_station_location & (ds[key_tweets.time_stations] == time)
            ] = tp_match.values[0]
        if n_time_not_found > 0:
            logging.info(f"Couldn't find {n_time_not_found}/{n_times_stations} for station.")
    logging.info(f"Couldn't find {n_time_not_found_total}/{ds.index.shape[0]} time values in total")
    return ds


def _assign_coordinates_of_nearest_station_to_tweets(ds, df_stations, key_tweets, coords, processes):
    tweet_latitudes = ds[key_tweets.latitude].values
    tweet_longitudes = ds[key_tweets.longitude].values

    weather_station_coordinates = unique_coordinates(df_stations)

    weather_station_latitudes = weather_station_coordinates[0, :]
    weather_station_longitudes = weather_station_coordinates[1, :]
    latitudes, longitudes, distances = find_closest_coordinate_and_distance(
        tweet_latitudes,
        tweet_longitudes,
        weather_station_latitudes,
        weather_station_longitudes,
        processes=processes,
    )
    ds[key_tweets.latitude_stations] = (coords, latitudes)
    ds[key_tweets.longitude_stations] = (coords, longitudes)
    ds[key_tweets.distance_stations] = (coords, distances)


def _get_station_from_lat_long(df_stations, latitude, longitude):
    mask_df_station_location = (df_stations["latitude"] == latitude) & (df_stations["longitude"] == longitude)
    df_single_station = df_stations.loc[mask_df_station_location]
    return df_single_station


def _increment_debug_counters(n_time_not_found_total, n_time_not_found):
    n_time_not_found_total += 1
    n_time_not_found += 1


def _initialize_debug_counters():
    n_time_not_found = 0
    n_times_stations = 0
    return n_time_not_found, n_times_stations


def add_station_number(
    df_stations: pd.DataFrame,
    key_longitude: str = "longitude",
    key_latitude: str = "latitude",
    key_station_number: str = "station_number",
):
    """Number weather stations based on unique longitude/latitude pair
    Add number as new column to df_stations"""
    station_number = 0
    df_stations[key_station_number] = np.full_like(df_stations["longitude"].values, np.nan)
    for latitude, longitude in zip(*unique_coordinates(df_stations)):
        mask_df_station_location = (df_stations[key_latitude] == latitude) & (df_stations[key_longitude] == longitude)
        df_stations.loc[mask_df_station_location, key_station_number] = station_number
        station_number += 1
    return df_stations


def get_counts_station(df_stations: pd.DataFrame, key_station_number: str = "station_number"):
    """Return station number and respective number of occurences"""
    value_counts = df_stations[key_station_number].value_counts()
    station_numbers = value_counts.index.values
    counts = value_counts.values
    return station_numbers, counts


def get_time_series_from_station_number(
    df_stations: pd.DataFrame,
    station_number: int,
    key_time: str = "ob_end_time",
    key_station_number: str = "station_number",
    key_tp: str = "prcp_amt",
):
    """Returns precipitation time series for station based on station number"""
    mask = df_stations[key_station_number] == station_number
    time, tp = df_stations[key_time].loc[mask].values, df_stations[key_tp].loc[mask].values
    mask_sort = np.argsort(time)  # type: ignore
    return time[mask_sort], tp[mask_sort]


def get_dataframe_from_station_number(
    df_stations: pd.DataFrame, station_number: int, key_station_number: str = "station_number"
):
    """Returns dataframe containing only data for specific station number"""
    mask = df_stations[key_station_number] == station_number
    return df_stations[mask]



import logging as logging_std_library

import a2.training.benchmarks as timer
import torch
import transformers
from transformers.trainer import *  # Ugly but probably needed ... # noeq


class TimerCallback(transformers.TrainerCallback):
    """
    # Example usage:
    tmr = timer.CPUGPUTimer()
    trainer = transformers.Trainer(
        # Other args...
        callbacks=[TimerCallback(tmr, gpu=True)]
    )
    trainer.train()
    tmr.print_all_time_stats()
    """

    def __init__(self, timer, gpu=False):
        super().__init__()
        self.timer = timer
        self.gpu = gpu

    def on_epoch_begin(self, args, state, control, **kwargs):
        self.timer.start(timer.TimeType.EPOCH)

    def on_epoch_end(self, args, state, control, **kwargs):
        self.timer.end(timer.TimeType.EPOCH)
        self.timer.complete_all()

    def on_step_begin(self, args, state, control, **kwargs):
        logging_std_library.info(
            f"Epoch {int(state.epoch)}: Start iteration step {state.global_step}/{state.max_steps} of training..."
        )
        self.timer.start(timer.TimeType.BATCH, gpu=self.gpu)

    def on_step_end(self, args, state, control, **kwargs):
        self.timer.end(timer.TimeType.BATCH, gpu=self.gpu)
        if state.global_step % 10 == 0:
            self.timer.complete_all()


class TimeLoaderWrapper:
    """Wrapper around a DataLoader (*not* a Dataset!) for I/O timing."""

    def __init__(self, loader, timer):
        self.loader = loader
        self.tmr = timer

    @staticmethod
    def time_loader(loader, tmr):
        if len(loader) > 0:
            tmr.start(timer.TimeType.IO)
        for i, data in enumerate(loader):
            tmr.end(timer.TimeType.IO)
            if i % 10 == 0:
                tmr.complete_all()
            yield data
            if i != len(loader) - 1:
                tmr.start(timer.TimeType.IO)

    def __iter__(self):
        return TimeLoaderWrapper.time_loader(self.loader, self.tmr)

    def __len__(self):
        return len(self.loader)

    def reset(self):
        if hasattr(self.loader, "reset"):
            self.loader.reset()


class TrainerWithTimer(Trainer):
    """
    Custom Trainer subclass to support finer-grained timing.
    Should also use the callback above.
    Adapted from original:
    # https://github.com/huggingface/transformers/blob/v4.25.1/src/transformers/trainer.py
    """

    def __init__(self, *args, **kwargs):
        [self.timer_callback] = kwargs["callbacks"]
        self.tmr = self.timer_callback.timer
        self.tmr_gpu = self.timer_callback.gpu
        super().__init__(*args, **kwargs)

    def training_step(self, model, inputs):
        model.train()
        inputs = self._prepare_inputs(inputs)

        if is_sagemaker_mp_enabled():
            loss_mb = smp_forward_backward(model, inputs, self.args.gradient_accumulation_steps)
            return loss_mb.reduce_mean().detach().to(self.args.device)

        if self.tmr:
            self.tmr.start(timer.TimeType.FORWARD, gpu=self.tmr_gpu)
        with self.compute_loss_context_manager():
            loss = self.compute_loss(model, inputs)

        if self.args.n_gpu > 1:
            loss = loss.mean()  # mean() to average on multi-gpu parallel training

        if self.args.gradient_accumulation_steps > 1 and not self.deepspeed:
            # deepspeed handles loss scaling by gradient_accumulation_steps in its `backward`
            loss = loss / self.args.gradient_accumulation_steps
        if self.tmr:
            self.tmr.end(timer.TimeType.FORWARD, gpu=self.tmr_gpu)
            self.tmr.start(timer.TimeType.BACKWARD, gpu=self.tmr_gpu)

        if self.do_grad_scaling:
            self.scaler.scale(loss).backward()
        elif self.use_apex:
            with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                scaled_loss.backward()
        elif self.deepspeed:
            # loss gets scaled under gradient_accumulation_steps in deepspeed
            loss = self.deepspeed.backward(loss)
        else:
            loss.backward()

        if self.tmr:
            self.tmr.end(timer.TimeType.BACKWARD, gpu=self.tmr_gpu)

        return loss.detach()

    def get_train_dataloader(self):
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")

        train_dataset = self.train_dataset
        data_collator = self.data_collator
        if is_datasets_available() and isinstance(train_dataset, datasets.Dataset):
            train_dataset = self._remove_unused_columns(train_dataset, description="training")
        else:
            data_collator = self._get_collator_with_removed_columns(data_collator, description="training")

        if isinstance(train_dataset, torch.utils.data.IterableDataset):
            if self.args.world_size > 1:
                train_dataset = IterableDatasetShard(
                    train_dataset,
                    batch_size=self._train_batch_size,
                    drop_last=self.args.dataloader_drop_last,
                    num_processes=self.args.world_size,
                    process_index=self.args.process_index,
                )

            dl = DataLoader(
                train_dataset,
                batch_size=self.args.per_device_train_batch_size,
                collate_fn=data_collator,
                num_workers=self.args.dataloader_num_workers,
                pin_memory=self.args.dataloader_pin_memory,
            )
        else:
            train_sampler = self._get_train_sampler()

            dl = DataLoader(
                train_dataset,
                batch_size=self._train_batch_size,
                sampler=train_sampler,
                collate_fn=data_collator,
                drop_last=self.args.dataloader_drop_last,
                num_workers=self.args.dataloader_num_workers,
                pin_memory=self.args.dataloader_pin_memory,
                worker_init_fn=seed_worker,
            )

        if self.tmr:
            return TimeLoaderWrapper(dl, self.tmr)
        return dl

    def save_model(self, output_dir: Optional[str] = None, _internal_call: bool = False):
        self.tmr.start(timer.TimeType.SAVING_MODEL)
        super().save_model(output_dir=output_dir, _internal_call=_internal_call)
        self.tmr.end(timer.TimeType.SAVING_MODEL)



import dataclasses
import logging
import resource
import time

import numpy as np


@dataclasses.dataclass(frozen=True)
class TimeType:
    EPOCH: str = "EPOCH"
    BATCH: str = "BATCH"
    FORWARD: str = "FORWARD"
    BACKWARD: str = "BACKWARD"
    IO: str = "IO"
    TRAINING: str = "TRAINING"
    RUN: str = "RUN"
    SAVING_MODEL: str = "SAVING_MODEL"


MAX_LENGTH_TIME_TYPES = max([len(x) for x in TimeType().__dict__.values()])


def current_time():
    return time.time()


class Timer:
    def __init__(self, print_all_single_time_stats=True, debug=True):
        self.times_archive = {}
        self.times_running = {}
        self.TimeType = TimeType()
        self.debug = debug
        self.print_all_single_time_stats = print_all_single_time_stats

    def start(self, time_type, gpu=None):
        if time_type in self.TimeType.__dict__.keys():
            if time_type in self.times_running:
                logging.warning(f"Overwriting unfinished timing of {time_type}!")
            self.times_running[time_type] = current_time()

    def _archive_timing(self, time_type, duration):
        if time_type not in self.times_archive:
            self.times_archive[time_type] = []
        self.times_archive[time_type].extend([duration])

    def end(self, time_type, gpu=None):
        if time_type in self.times_running:
            duration = current_time() - self.times_running.pop(time_type)
            self._archive_timing(time_type, duration=duration)
            if self.print_all_single_time_stats:
                logging.info(f"{time_type:<10}: took {duration}")
        else:
            logging.warning(f"Attempting to finish timer type {time_type}, which was never started!")

    def complete_all(self):
        # for _type in self.times_running.copy().keys():
        #     self.end(_type)
        pass

    def print_all_time_stats(self):
        for _type, times in self.times_archive.items():
            print(
                f"{_type:<15}; "
                f"min: {min(times):.04e}, "
                f"max: {max(times):.04e}, "
                f"mean: {np.mean(times):.04e}, "
                f"median: {np.median(times):.04e}, "
                f"counts: {len(times):>5}"
            )
            if self.debug:
                for _type, elapsed_time in self.times_running.items():
                    logging.info(f"{_type} still running for {elapsed_time}s.")


def get_max_memory_usage():
    """In bytes"""
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1000



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['a2',
 'a2.cli',
 'a2.cli.cli_plotting',
 'a2.data',
 'a2.data.emoji',
 'a2.data.vocabularies',
 'a2.dataset',
 'a2.plotting',
 'a2.preprocess',
 'a2.training',
 'a2.twitter',
 'a2.utils']

package_data = \
{'': ['*']}

install_requires = \
['DateTime>=4.4,<5.0',
 'Shapely>=1.8.2,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'convertbng>=0.6.39,<0.7.0',
 'dask>=2022.6.1,<2023.0.0',
 'emoji<1.0.0',
 'geopy>=2.3.0,<3.0.0',
 'h5netcdf>=1.0.1,<2.0.0',
 'h5py>=3.7.0,<4.0.0',
 'ipython>=8.7.0,<9.0.0',
 'ipywidgets>=7.7.1,<8.0.0',
 'jsonlines>=3.1.0,<4.0.0',
 'jupyterlab>=3.4.3,<4.0.0',
 'kaleido==0.2.1',
 'mantik>=0.1.0,<0.2.0',
 'matplotlib>=3.5.2,<4.0.0',
 'netCDF4>=1.6.0,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'plotly>=5.11.0,<6.0.0',
 'pyproj>=3.3.1,<4.0.0',
 'pytest-recording>=0.12.1,<0.13.0',
 'pytest-xdist>=2.5.0,<3.0.0',
 'rasterio>=1.3.3,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'rioxarray>=0.12.2,<0.13.0',
 'scikit-learn>=1.1.2,<2.0.0',
 'seaborn>=0.12.1,<0.13.0',
 'spacymoji>=3.0.1,<4.0.0',
 'tweepy>=4.10.0,<5.0.0',
 'urllib3>=1.26.12,<2.0.0',
 'wget>=3.2,<4.0',
 'wordcloud>=1.8.2,<2.0.0',
 'xarray>=2022.3.0,<2023.0.0']

entry_points = \
{'console_scripts': ['a2 = a2.cli.main:cli']}

setup_kwargs = {
    'name': 'a2',
    'version': '0.2.10',
    'description': 'Package for predicting information about the weather from social media data as application 2 for maelstrom project',
    'long_description': 'None',
    'author': 'Kristian Ehlert',
    'author_email': 'kristian.ehlert@4-cast.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)
